from collections.abc import Callable, Iterator, Sequence
from re import Pattern
from types import ModuleType
from typing import Any, TypeAlias, overload

from django.core.checks.messages import CheckMessage
from django.http import HttpRequest, HttpResponse
from django.urls import _AnyURL
from django.utils.datastructures import MultiValueDict
from django.utils.functional import cached_property

class ResolverMatch:
    func: Callable[..., Any]
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    url_name: str | None
    app_names: list[str]
    app_name: str
    namespaces: list[str]
    namespace: str
    view_name: str
    route: str
    tried: Any | None
    _func_path: str
    def __init__(
        self,
        func: Callable[..., Any],
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
        url_name: str | None = None,
        app_names: list[str | None] | None = None,
        namespaces: list[str | None] | None = None,
        route: str | None = None,
        tried: Any | None = None,
        captured_kwargs: dict[str, Any] | None = None,
        extra_kwargs: dict[str, Any] | None = None,
    ) -> None: ...
    def __getitem__(self, index: int) -> Any: ...
    # for tuple unpacking -- python/mypy#2220
    def __iter__(self) -> Iterator[Any]: ...

def get_resolver(urlconf: str | None = None) -> URLResolver: ...
def get_ns_resolver(ns_pattern: str, resolver: URLResolver, converters: tuple[Any, ...]) -> URLResolver: ...

_Pattern: TypeAlias = RegexPattern | RoutePattern | LocalePrefixPattern

class LocaleRegexDescriptor:
    @overload
    def __get__(self, instance: None, cls: type[_Pattern] | None = None) -> LocaleRegexDescriptor: ...
    @overload
    def __get__(self, instance: _Pattern, cls: type[_Pattern] | None = None) -> Pattern[str]: ...

class CheckURLMixin:
    def describe(self) -> str: ...

class RegexPattern(CheckURLMixin):
    regex: LocaleRegexDescriptor
    name: str | None
    converters: dict[str, Any]
    def __init__(self, regex: str, name: str | None = None, is_endpoint: bool = False) -> None: ...
    def match(self, path: str) -> tuple[str, tuple[Any, ...], dict[str, str]] | None: ...
    def check(self) -> list[CheckMessage]: ...

whitespace_set: frozenset[str]

class LocaleRegexRouteDescriptor:
    def __get__(self, instance: RoutePattern | None, cls: type | None = None) -> Pattern[str]: ...

class RoutePattern(CheckURLMixin):
    regex: LocaleRegexRouteDescriptor
    name: str | None
    converters: dict[str, Any]
    def __init__(self, route: str, name: str | None = None, is_endpoint: bool = False) -> None: ...
    def match(self, path: str) -> tuple[str, tuple[Any, ...], dict[str, int | str]] | None: ...
    def check(self) -> list[CheckMessage]: ...

class LocalePrefixPattern:
    prefix_default_language: bool
    converters: dict[str, Any]
    def __init__(self, prefix_default_language: bool = True) -> None: ...
    @property
    def regex(self) -> Pattern[str]: ...
    @property
    def language_prefix(self) -> str: ...
    def match(self, path: str) -> tuple[str, tuple[Any, ...], dict[str, Any]] | None: ...
    def check(self) -> list[CheckMessage]: ...
    def describe(self) -> str: ...

class URLPattern:
    pattern: _Pattern
    callback: Callable[..., Any]
    default_args: dict[str, Any]
    name: str | None
    def __init__(
        self,
        pattern: _Pattern,
        callback: Callable[..., Any],
        default_args: dict[str, Any] | None = None,
        name: str | None = None,
    ) -> None: ...
    def check(self) -> list[CheckMessage]: ...
    def resolve(self, path: str) -> ResolverMatch | None: ...
    @cached_property
    def lookup_str(self) -> str: ...

class URLResolver:
    pattern: _Pattern
    urlconf_name: str | Sequence[_AnyURL] | ModuleType
    callback: None
    default_kwargs: dict[str, Any]
    namespace: str | None
    app_name: str | None
    _local: Any
    _reverse_dict: MultiValueDict
    def __init__(
        self,
        pattern: _Pattern,
        urlconf_name: str | Sequence[_AnyURL] | ModuleType,
        default_kwargs: dict[str, Any] | None = None,
        app_name: str | None = None,
        namespace: str | None = None,
    ) -> None: ...
    @property
    def reverse_dict(self) -> MultiValueDict: ...
    @property
    def namespace_dict(self) -> dict[str, tuple[str, URLResolver]]: ...
    @property
    def app_dict(self) -> dict[str, list[str]]: ...
    @cached_property
    def urlconf_module(self) -> ModuleType | None | Sequence[_AnyURL]: ...
    @cached_property
    def url_patterns(self) -> list[_AnyURL]: ...
    def check(self) -> list[CheckMessage]: ...
    def _populate(self) -> None: ...
    def _is_callback(self, name: str) -> bool: ...
    def resolve(self, path: str) -> ResolverMatch: ...
    def resolve_error_handler(self, view_type: int) -> Callable[[HttpRequest], HttpResponse]: ...
    def reverse(self, lookup_view: str, *args: Any, **kwargs: Any) -> str: ...



import datetime
from collections.abc import Iterator, Mapping, Sequence
from contextlib import AbstractContextManager
from decimal import Decimal
from logging import Logger
from types import TracebackType
from typing import Any, Literal, Protocol, TypeAlias, overload, type_check_only
from uuid import UUID

from typing_extensions import Self

logger: Logger

# Protocol matching psycopg2.sql.Composable, to avoid depending psycopg2
@type_check_only
class _Composable(Protocol):
    def as_string(self, context: Any, /) -> str: ...
    def __add__(self, other: Self, /) -> _Composable: ...
    def __mul__(self, n: int, /) -> _Composable: ...

_ExecuteQuery: TypeAlias = str | _Composable

# Python types that can be adapted to SQL.
_SQLType: TypeAlias = (
    None
    | bool
    | int
    | float
    | Decimal
    | str
    | bytes
    | datetime.date
    | datetime.datetime
    | UUID
    | tuple[Any, ...]
    | list[Any]
)
_ExecuteParameters: TypeAlias = Sequence[_SQLType] | Mapping[str, _SQLType] | None

class CursorWrapper:
    cursor: Any
    db: Any
    def __init__(self, cursor: Any, db: Any) -> None: ...
    WRAP_ERROR_ATTRS: Any
    APPS_NOT_READY_WARNING_MSG: str

    def __getattr__(self, attr: str) -> Any: ...
    def __iter__(self) -> Iterator[tuple[Any, ...]]: ...
    def __enter__(self) -> Self: ...
    def __exit__(
        self,
        type: type[BaseException] | None,
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None: ...
    def callproc(
        self, procname: str, params: Sequence[Any] | None = None, kparams: dict[str, int] | None = None
    ) -> Any: ...
    def execute(self, sql: _ExecuteQuery, params: _ExecuteParameters | None = None) -> Any: ...
    def executemany(self, sql: _ExecuteQuery, param_list: Sequence[_ExecuteParameters]) -> Any: ...

class CursorDebugWrapper(CursorWrapper):
    def debug_sql(
        self,
        sql: str | None = None,
        params: _ExecuteParameters | Sequence[_ExecuteParameters] | None = None,
        use_last_executed_query: bool = False,
        many: bool = False,
    ) -> AbstractContextManager[None]: ...

def debug_transaction(connection: Any, sql: str) -> AbstractContextManager[None]: ...
def split_tzname_delta(tzname: str) -> tuple[str, str | None, str | None]: ...
@overload
def typecast_date(s: None | Literal[""]) -> None: ...  # type: ignore[overload-overlap]
@overload
def typecast_date(s: str) -> datetime.date: ...
@overload
def typecast_time(s: None | Literal[""]) -> None: ...  # type: ignore[overload-overlap]
@overload
def typecast_time(s: str) -> datetime.time: ...
@overload
def typecast_timestamp(s: None | Literal[""]) -> None: ...  # type: ignore[overload-overlap]
@overload
def typecast_timestamp(s: str) -> datetime.datetime: ...
def split_identifier(identifier: str) -> tuple[str, str]: ...
def truncate_name(identifier: str, length: int | None = None, hash_len: int = 4) -> str: ...
def names_digest(*args: str, length: int) -> str: ...
def format_number(value: Decimal | None, max_digits: int | None, decimal_places: int | None) -> str | None: ...
def strip_quotes(table_name: str) -> str: ...



import enum
import sys
from typing import Any, Literal, TypeVar, overload, type_check_only

from _typeshed import ConvertibleToInt
from django.utils.functional import _StrOrPromise
from typing_extensions import override

if sys.version_info >= (3, 11):
    from enum import EnumType, IntEnum, StrEnum
    from enum import property as enum_property
else:
    from enum import EnumMeta as EnumType
    from types import DynamicClassAttribute as enum_property

    class ReprEnum(enum.Enum): ...  # type: ignore[misc]
    class IntEnum(int, ReprEnum): ...  # type: ignore[misc]
    class StrEnum(str, ReprEnum): ...  # type: ignore[misc]

_Self = TypeVar("_Self", bound=ChoicesType)

class ChoicesType(EnumType):
    __empty__: _StrOrPromise
    def __new__(
        metacls: type[_Self], classname: str, bases: tuple[type, ...], classdict: enum._EnumDict, **kwds: Any
    ) -> _Self: ...
    @property
    def names(self) -> list[str]: ...
    @property
    def choices(self) -> list[tuple[Any, _StrOrPromise]]: ...
    @property
    def labels(self) -> list[_StrOrPromise]: ...
    @property
    def values(self) -> list[Any]: ...
    if sys.version_info < (3, 12):
        @override
        def __contains__(self, member: Any) -> bool: ...

class Choices(enum.Enum, metaclass=ChoicesType):  # type: ignore[misc]
    _label_: _StrOrPromise
    do_not_call_in_templates: Literal[True]

    @enum_property
    def label(self) -> _StrOrPromise: ...
    @enum_property
    @override
    def value(self) -> Any: ...

# fake, to keep simulate class properties
@type_check_only
class _IntegerChoicesType(ChoicesType):
    @property
    @override
    def choices(self) -> list[tuple[int, _StrOrPromise]]: ...
    @property
    @override
    def values(self) -> list[int]: ...

# In reality, the `__init__` overloads provided below should also support
# all the arguments of `int.__new__`/`str.__new__` (e.g. `base`, `encoding`).
# They are omitted on purpose to avoid having convoluted stubs for these enums:
class IntegerChoices(Choices, IntEnum, metaclass=_IntegerChoicesType):  # type: ignore[misc]
    @overload
    def __init__(self, x: ConvertibleToInt) -> None: ...
    @overload
    def __init__(self, x: ConvertibleToInt, label: _StrOrPromise) -> None: ...
    @enum_property
    @override
    def value(self) -> int: ...

# fake, to keep simulate class properties
@type_check_only
class _TextChoicesType(ChoicesType):
    @property
    @override
    def choices(self) -> list[tuple[str, _StrOrPromise]]: ...
    @property
    @override
    def values(self) -> list[str]: ...

class TextChoices(Choices, StrEnum, metaclass=_TextChoicesType):  # type: ignore[misc]
    @overload
    def __init__(self, object: str) -> None: ...
    @overload
    def __init__(self, object: str, label: _StrOrPromise) -> None: ...
    @enum_property
    @override
    def value(self) -> str: ...

__all__ = ["Choices", "IntegerChoices", "TextChoices"]



from mypy.nodes import TypeInfo
from mypy.plugin import ClassDefContext

from mypy_django_plugin.lib import fullnames, helpers


def make_meta_nested_class_inherit_from_any(ctx: ClassDefContext) -> None:
    meta_node = helpers.get_nested_meta_node_for_current_class(ctx.cls.info)
    if meta_node is None:
        if not ctx.api.final_iteration:
            ctx.api.defer()
    else:
        meta_node.fallback_to_any = True


def transform_form_class(ctx: ClassDefContext) -> None:
    sym = ctx.api.lookup_fully_qualified_or_none(fullnames.BASEFORM_CLASS_FULLNAME)
    if sym is not None and isinstance(sym.node, TypeInfo):
        bases = helpers.get_django_metadata_bases(sym.node, "baseform_bases")
        bases[ctx.cls.fullname] = 1

    make_meta_nested_class_inherit_from_any(ctx)



from typing import Any, NamedTuple

from django.db.backends.base.introspection import BaseDatabaseIntrospection
from django.db.backends.sqlite3.base import DatabaseWrapper

class FieldInfo(NamedTuple):
    name: str
    type_code: int
    display_size: int | None
    internal_size: int | None
    precision: int | None
    scale: int | None
    null_ok: bool
    default: str | None
    collation: str | None
    pk: bool
    has_json_constraint: bool

field_size_re: Any

def get_field_size(name: str) -> int | None: ...

class FlexibleFieldLookupDict:
    base_data_types_reverse: Any
    def __getitem__(self, key: str) -> Any: ...

class DatabaseIntrospection(BaseDatabaseIntrospection):
    connection: DatabaseWrapper



from django.contrib.admin.options import ModelAdmin
from django.contrib.admin.sites import AdminSite
from django.db.models import Model
from django.db.models.fields.related import ForeignKey
from django.http.request import HttpRequest
from django.views.generic.list import BaseListView

class AutocompleteJsonView(BaseListView):
    admin_site: AdminSite | None
    model_admin: ModelAdmin
    source_field: ForeignKey
    term: str
    def serialize_result(self, obj: Model, to_field_name: str) -> dict[str, str]: ...
    def process_request(self, request: HttpRequest) -> tuple[str, ModelAdmin, ForeignKey, str]: ...
    def has_perm(self, request: HttpRequest, obj: Model | None = ...) -> bool: ...



from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Generic, TypeVar, overload

from django.db.models.enums import TextChoices as TextChoices
from django.utils.module_loading import import_string as import_string
from django.utils.translation import pgettext_lazy as pgettext_lazy
from typing_extensions import ParamSpec

from .backends.base import BaseTaskBackend
from .exceptions import TaskResultMismatch as TaskResultMismatch

DEFAULT_TASK_BACKEND_ALIAS: str
DEFAULT_TASK_PRIORITY: int
DEFAULT_TASK_QUEUE_NAME: str
TASK_MAX_PRIORITY: int
TASK_MIN_PRIORITY: int
TASK_REFRESH_ATTRS: set[str]

class TaskResultStatus(TextChoices):
    READY = ...
    RUNNING = ...
    FAILED = ...
    SUCCESSFUL = ...

_P = ParamSpec("_P")
_R = TypeVar("_R")

@dataclass(kw_only=True)
class Task(Generic[_P, _R]):
    priority: int
    func: Callable[_P, _R]
    backend: str
    queue_name: str
    run_after: datetime | None
    takes_context: bool = ...
    def __post_init__(self) -> None: ...
    @property
    def name(self) -> str: ...
    def using(
        self,
        *,
        priority: int | None = None,
        queue_name: str | None = None,
        run_after: datetime | None = None,
        backend: str | None = None,
    ) -> Task[_P, _R]: ...
    def enqueue(self, *args: _P.args, **kwargs: _P.kwargs) -> TaskResult[_P, _R]: ...
    async def aenqueue(self, *args: _P.args, **kwargs: _P.kwargs) -> TaskResult[_P, _R]: ...
    def get_result(self, result_id: str) -> TaskResult[_P, _R]: ...
    async def aget_result(self, result_id: str) -> TaskResult[_P, _R]: ...
    def call(self, *args: _P.args, **kwargs: _P.kwargs) -> _R: ...
    async def acall(self, *args: _P.args, **kwargs: _P.kwargs) -> _R: ...
    def get_backend(self) -> BaseTaskBackend: ...
    @property
    def module_path(self) -> str: ...

@overload
def task(
    function: Callable[_P, _R],
    *,
    priority: int | None = None,
    queue_name: str | None = None,
    backend: str | None = None,
    takes_context: bool = ...,
) -> Task[_P, _R]: ...
@overload
def task(
    *,
    priority: int | None = None,
    queue_name: str | None = None,
    backend: str | None = None,
    takes_context: bool = ...,
) -> Callable[[Callable[_P, _R]], Task[_P, _R]]: ...

@dataclass(kw_only=True)
class TaskError:
    exception_class_path: str
    traceback: str
    @property
    def exception_class(self) -> type[BaseException]: ...

@dataclass(kw_only=True)
class TaskResult(Generic[_P, _R]):
    task: Task[_P, _R]
    id: str
    status: TaskResultStatus
    enqueued_at: datetime | None
    started_at: datetime | None
    finished_at: datetime | None
    last_attempted_at: datetime | None
    args: list[Any]
    kwargs: dict[str, Any]
    backend: str
    errors: list[TaskError]
    worker_ids: list[str]
    def __post_init__(self) -> None: ...
    @property
    def return_value(self) -> _R: ...
    @property
    def is_finished(self) -> bool: ...
    @property
    def attempts(self) -> int: ...
    def refresh(self) -> None: ...
    async def arefresh(self) -> None: ...

@dataclass(kw_only=True)
class TaskContext(Generic[_P, _R]):
    task_result: TaskResult[_P, _R]
    @property
    def attempt(self) -> int: ...



from collections.abc import Sequence
from typing import Any, ClassVar

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models.expressions import Combinable, Func
from django.db.models.fields import IntegerField
from django.db.models.functions.mixins import FixDurationInputMixin, NumericOutputFieldMixin
from django.db.models.query import _OrderByFieldName
from django.db.models.sql.compiler import SQLCompiler, _AsSqlType
from typing_extensions import override

class Aggregate(Func):
    name: str | None
    filter: Any
    allow_distinct: bool
    allow_order_by: bool
    empty_result_set_value: int | None
    def __init__(
        self,
        *expressions: Any,
        distinct: bool = False,
        filter: Any | None = None,
        default: Any | None = None,
        order_by: Sequence[_OrderByFieldName] | None = None,
        **extra: Any,
    ) -> None: ...
    @property
    def default_alias(self) -> str: ...
    @override
    def as_sql(  # type: ignore[override]
        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any
    ) -> _AsSqlType: ...

class AnyValue(Aggregate):
    @override
    def as_sql(  # type: ignore[override]
        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any
    ) -> _AsSqlType: ...

class Avg(FixDurationInputMixin, NumericOutputFieldMixin, Aggregate): ...

class Count(Aggregate):
    output_field: ClassVar[IntegerField]
    def __init__(self, expression: Any, filter: Any | None = None, **extra: Any) -> None: ...

class Max(Aggregate): ...
class Min(Aggregate): ...

class StdDev(NumericOutputFieldMixin, Aggregate):
    def __init__(self, expression: Any, sample: bool = False, **extra: Any) -> None: ...

class StringAgg(Aggregate):
    def __init__(self, expression: Combinable | str, delimiter: str | Combinable, **extra: Any) -> None: ...
    def as_oracle(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any) -> _AsSqlType: ...
    def as_mysql(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any) -> _AsSqlType: ...
    @override
    def as_sqlite(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any) -> _AsSqlType: ...

class Sum(FixDurationInputMixin, Aggregate): ...

class Variance(NumericOutputFieldMixin, Aggregate):
    def __init__(self, expression: Any, sample: bool = False, **extra: Any) -> None: ...

__all__ = ["Aggregate", "AnyValue", "Avg", "Count", "Max", "Min", "StdDev", "StringAgg", "Sum", "Variance"]



from collections.abc import Callable, Iterable
from datetime import date, datetime
from sqlite3 import dbapi2 as Database
from types import ModuleType
from typing import Any, TypeVar

from django.db.backends.base.base import BaseDatabaseWrapper
from typing_extensions import override

from .client import DatabaseClient
from .creation import DatabaseCreation
from .features import DatabaseFeatures
from .introspection import DatabaseIntrospection
from .operations import DatabaseOperations
from .schema import DatabaseSchemaEditor

_R = TypeVar("_R")

def decoder(conv_func: Callable[[str], _R]) -> Callable[[bytes], _R]: ...
def adapt_date(val: date) -> str: ...
def adapt_datetime(val: datetime) -> str: ...

class DatabaseWrapper(BaseDatabaseWrapper):
    client: DatabaseClient
    creation: DatabaseCreation
    features: DatabaseFeatures
    introspection: DatabaseIntrospection
    ops: DatabaseOperations
    operators: dict[str, str]

    pattern_esc: str
    pattern_ops: dict[str, str]
    transaction_modes: frozenset[str]
    Database: ModuleType
    SchemaEditorClass: type[DatabaseSchemaEditor]
    client_class: type[DatabaseClient]
    creation_class: type[DatabaseCreation]
    features_class: type[DatabaseFeatures]
    introspection_class: type[DatabaseIntrospection]
    ops_class: type[DatabaseOperations]

    def is_in_memory_db(self) -> bool: ...

FORMAT_QMARK_REGEX: Any

class SQLiteCursorWrapper(Database.Cursor):
    @override
    def execute(self, query: str, params: Iterable[Any] | None = None) -> SQLiteCursorWrapper: ...  # type: ignore[override]
    @override
    def executemany(self, query: str, param_list: Iterable[Iterable[Any]]) -> SQLiteCursorWrapper: ...  # type: ignore[override]
    def convert_query(self, query: str, *, param_names: Any = None) -> str: ...



import datetime
import sys
from collections.abc import AsyncIterator, Collection, Iterable, Iterator, Mapping, Sequence, Sized
from typing import Any, Generic, Literal, NamedTuple, TypeAlias, overload

from django.db.backends.utils import _ExecuteQuery
from django.db.models import Manager
from django.db.models.base import Model
from django.db.models.expressions import Combinable, OrderBy
from django.db.models.sql.query import Query, RawQuery
from django.utils.functional import cached_property
from typing_extensions import Self, TypeVar, override

_T = TypeVar("_T", covariant=True)
_Model = TypeVar("_Model", bound=Model, covariant=True)
_Row = TypeVar("_Row", covariant=True, default=_Model)  # ONLY use together with _Model
_TupleT = TypeVar("_TupleT", bound=tuple[Any, ...], covariant=True)

# The type of the lookup passed to Prefetch(...)
# This will be specialized to a `LiteralString` in the plugin for further processing and validation
_LookupT = TypeVar("_LookupT", bound=str, covariant=True)
# The type of the queryset passed to Prefetch(...)
_PrefetchedQuerySetT = TypeVar("_PrefetchedQuerySetT", bound=QuerySet[Model], covariant=True, default=QuerySet[Model])
# The attribute name to store the prefetched list[_PrefetchedQuerySetT]
# This will be specialized to a `LiteralString` in the plugin for further processing and validation
_ToAttrT = TypeVar("_ToAttrT", bound=str, covariant=True, default=str)

_OrderByFieldName: TypeAlias = str | Combinable

MAX_GET_RESULTS: int
REPR_OUTPUT_SIZE: int
PROHIBITED_FILTER_KWARGS: frozenset[str]

class BaseIterable(Generic[_T]):
    queryset: QuerySet[Model]
    chunked_fetch: bool
    chunk_size: int
    def __init__(self, queryset: QuerySet[Model], chunked_fetch: bool = False, chunk_size: int = 100) -> None: ...
    def __aiter__(self) -> AsyncIterator[_T]: ...

class ModelIterable(BaseIterable[_Model], Generic[_Model]):
    def __iter__(self) -> Iterator[_Model]: ...

class RawModelIterable(BaseIterable[dict[str, Any]]):
    def __iter__(self) -> Iterator[dict[str, Any]]: ...

class ValuesIterable(BaseIterable[dict[str, Any]]):
    def __iter__(self) -> Iterator[dict[str, Any]]: ...

class ValuesListIterable(BaseIterable[_TupleT]):
    def __iter__(self) -> Iterator[_TupleT]: ...

class NamedValuesListIterable(ValuesListIterable[NamedTuple]):
    @override
    def __iter__(self) -> Iterator[NamedTuple]: ...

class FlatValuesListIterable(BaseIterable[_T]):
    def __iter__(self) -> Iterator[_T]: ...

class QuerySet(Iterable[_Row], Sized, Generic[_Model, _Row]):
    model: type[_Model]
    query: Query
    _iterable_class: type[BaseIterable]
    _result_cache: list[_Row] | None
    def __init__(
        self,
        model: type[Model] | None = None,
        query: Query | None = None,
        using: str | None = None,
        hints: dict[str, Model] | None = None,
    ) -> None: ...
    @classmethod
    def as_manager(cls) -> Manager[_Model]: ...
    def __deepcopy__(self, memo: dict[int, Any]) -> Self: ...
    if sys.version_info >= (3, 11):
        @override
        def __getstate__(self) -> dict[str, Any]: ...
    else:
        def __getstate__(self) -> dict[str, Any]: ...
    @override
    def __len__(self) -> int: ...
    @override
    def __iter__(self) -> Iterator[_Row]: ...
    def __aiter__(self) -> AsyncIterator[_Row]: ...
    def __bool__(self) -> bool: ...
    @overload
    def __getitem__(self, i: int) -> _Row: ...
    @overload
    def __getitem__(self, s: slice) -> Self: ...
    def __class_getitem__(cls, item: type[_Model]) -> Self: ...
    # Technically, the other QuerySet must be of the same type _T, but _T is covariant
    def __and__(self, other: QuerySet[_Model, _Row]) -> Self: ...
    def __or__(self, other: QuerySet[_Model, _Row]) -> Self: ...
    def __xor__(self, other: QuerySet[_Model, _Row]) -> Self: ...
    # IMPORTANT: When updating any of the following methods' signatures, please ALSO modify
    # the corresponding method in BaseManager.
    def iterator(self, chunk_size: int | None = None) -> Iterator[_Row]: ...
    def aiterator(self, chunk_size: int = 2000) -> AsyncIterator[_Row]: ...
    def aggregate(self, *args: Any, **kwargs: Any) -> dict[str, Any]: ...
    async def aaggregate(self, *args: Any, **kwargs: Any) -> dict[str, Any]: ...
    def count(self) -> int: ...
    async def acount(self) -> int: ...
    def get(self, *args: Any, **kwargs: Any) -> _Row: ...
    async def aget(self, *args: Any, **kwargs: Any) -> _Row: ...
    def create(self, **kwargs: Any) -> _Model: ...
    async def acreate(self, **kwargs: Any) -> _Model: ...
    def bulk_create(
        self,
        objs: Iterable[_Model],
        batch_size: int | None = None,
        ignore_conflicts: bool = False,
        update_conflicts: bool = False,
        update_fields: Collection[str] | None = None,
        unique_fields: Collection[str] | None = None,
    ) -> list[_Model]: ...
    async def abulk_create(
        self,
        objs: Iterable[_Model],
        batch_size: int | None = None,
        ignore_conflicts: bool = False,
        update_conflicts: bool = False,
        update_fields: Collection[str] | None = None,
        unique_fields: Collection[str] | None = None,
    ) -> list[_Model]: ...
    def bulk_update(self, objs: Iterable[_Model], fields: Iterable[str], batch_size: int | None = None) -> int: ...
    async def abulk_update(
        self, objs: Iterable[_Model], fields: Iterable[str], batch_size: int | None = None
    ) -> int: ...
    def get_or_create(self, defaults: Mapping[str, Any] | None = None, **kwargs: Any) -> tuple[_Model, bool]: ...
    async def aget_or_create(self, defaults: Mapping[str, Any] | None = None, **kwargs: Any) -> tuple[_Model, bool]: ...
    def update_or_create(
        self,
        defaults: Mapping[str, Any] | None = None,
        create_defaults: Mapping[str, Any] | None = None,
        **kwargs: Any,
    ) -> tuple[_Model, bool]: ...
    async def aupdate_or_create(
        self,
        defaults: Mapping[str, Any] | None = None,
        create_defaults: Mapping[str, Any] | None = None,
        **kwargs: Any,
    ) -> tuple[_Model, bool]: ...
    def earliest(self, *fields: str | OrderBy) -> _Row: ...
    async def aearliest(self, *fields: str | OrderBy) -> _Row: ...
    def latest(self, *fields: str | OrderBy) -> _Row: ...
    async def alatest(self, *fields: str | OrderBy) -> _Row: ...
    def first(self) -> _Row | None: ...
    async def afirst(self) -> _Row | None: ...
    def last(self) -> _Row | None: ...
    async def alast(self) -> _Row | None: ...
    def in_bulk(self, id_list: Iterable[Any] | None = None, *, field_name: str = "pk") -> dict[Any, _Model]: ...
    async def ain_bulk(self, id_list: Iterable[Any] | None = None, *, field_name: str = "pk") -> dict[Any, _Model]: ...
    def delete(self) -> tuple[int, dict[str, int]]: ...
    async def adelete(self) -> tuple[int, dict[str, int]]: ...
    def _raw_delete(self, using: str | None) -> int: ...
    def update(self, **kwargs: Any) -> int: ...
    async def aupdate(self, **kwargs: Any) -> int: ...
    def exists(self) -> bool: ...
    async def aexists(self) -> bool: ...
    def contains(self, obj: Model) -> bool: ...
    async def acontains(self, obj: Model) -> bool: ...
    def explain(self, *, format: str | None = None, **options: Any) -> str: ...
    async def aexplain(self, *, format: str | None = None, **options: Any) -> str: ...
    def raw(
        self,
        raw_query: _ExecuteQuery,
        params: Any = (),
        translations: dict[str, str] | None = None,
        using: str | None = None,
    ) -> RawQuerySet[_Model]: ...
    # The type of values may be overridden to be more specific in the mypy plugin, depending on the fields param
    def values(self, *fields: str | Combinable, **expressions: Any) -> QuerySet[_Model, dict[str, Any]]: ...
    # The type of values_list may be overridden to be more specific in the mypy plugin, depending on the fields param
    def values_list(
        self, *fields: str | Combinable, flat: bool = False, named: bool = False
    ) -> QuerySet[_Model, Any]: ...
    def dates(
        self,
        field_name: str,
        kind: Literal["year", "month", "week", "day"],
        order: Literal["ASC", "DESC"] = "ASC",
    ) -> QuerySet[_Model, datetime.date]: ...
    def datetimes(
        self,
        field_name: str,
        kind: Literal["year", "month", "week", "day", "hour", "minute", "second"],
        order: Literal["ASC", "DESC"] = "ASC",
        tzinfo: datetime.tzinfo | None = None,
    ) -> QuerySet[_Model, datetime.datetime]: ...
    def none(self) -> Self: ...
    def all(self) -> Self: ...
    def filter(self, *args: Any, **kwargs: Any) -> Self: ...
    def exclude(self, *args: Any, **kwargs: Any) -> Self: ...
    def complex_filter(self, filter_obj: Any) -> Self: ...
    def union(self, *other_qs: QuerySet[Model, Any], all: bool = False) -> Self: ...
    def intersection(self, *other_qs: QuerySet[Model, Any]) -> Self: ...
    def difference(self, *other_qs: QuerySet[Model, Any]) -> Self: ...
    def select_for_update(
        self, nowait: bool = False, skip_locked: bool = False, of: Sequence[str] = (), no_key: bool = False
    ) -> Self: ...
    @overload
    def select_related(self, clear: None, /) -> Self: ...
    @overload
    def select_related(self, *fields: str) -> Self: ...
    @overload
    def prefetch_related(self, clear: None, /) -> Self: ...
    @overload
    def prefetch_related(self, *lookups: str | Prefetch[_LookupT, _PrefetchedQuerySetT, _ToAttrT]) -> Self: ...
    def annotate(self, *args: Any, **kwargs: Any) -> Self: ...
    def alias(self, *args: Any, **kwargs: Any) -> Self: ...
    def order_by(self, *field_names: _OrderByFieldName) -> Self: ...
    def distinct(self, *field_names: str) -> Self: ...
    # extra() return type won't be supported any time soon
    def extra(
        self,
        select: dict[str, Any] | None = None,
        where: Sequence[str] | None = None,
        params: Sequence[Any] | None = None,
        tables: Sequence[str] | None = None,
        order_by: Sequence[_OrderByFieldName] | None = None,
        select_params: Sequence[Any] | None = None,
    ) -> QuerySet[Any, Any]: ...
    def reverse(self) -> Self: ...
    @overload
    def defer(self, clear: None, /) -> Self: ...
    @overload
    def defer(self, *fields: str) -> Self: ...
    def only(self, *fields: str) -> Self: ...
    def using(self, alias: str | None) -> Self: ...
    @property
    def ordered(self) -> bool: ...
    @property
    def db(self) -> str: ...
    def _fetch_all(self) -> None: ...
    def resolve_expression(self, *args: Any, **kwargs: Any) -> Any: ...

class RawQuerySet(Iterable[_Model], Sized):
    raw_query: RawQuery | str
    model: type[_Model] | None
    query: RawQuery
    params: tuple[Any, ...] | None
    translations: dict[str, str]

    def __init__(
        self,
        raw_query: RawQuery | str,
        model: type[_Model] | None = None,
        query: Query | None = None,
        params: tuple[Any, ...] = (),
        translations: dict[str, str] | None = None,
        using: str | None = None,
        hints: dict[str, Model] | None = None,
    ) -> None: ...
    def resolve_model_init_order(self) -> tuple[list[str], list[int], list[tuple[str, int]]]: ...
    @overload
    def prefetch_related(self, clear: None, /) -> Self: ...
    @overload
    def prefetch_related(self, *lookups: str | Prefetch[_LookupT, _PrefetchedQuerySetT, _ToAttrT]) -> Self: ...
    @override
    def __len__(self) -> int: ...
    def __bool__(self) -> bool: ...
    @override
    def __iter__(self) -> Iterator[_Model]: ...
    def __aiter__(self) -> AsyncIterator[_Model]: ...
    def iterator(self) -> Iterator[_Model]: ...
    @overload
    def __getitem__(self, k: int) -> _Model: ...
    @overload
    def __getitem__(self, k: str) -> Any: ...
    @overload
    def __getitem__(self, k: slice) -> Self: ...
    @property
    def db(self) -> str: ...
    def using(self, alias: str | None) -> Self: ...
    @cached_property
    def columns(self) -> list[str]: ...
    @cached_property
    def model_fields(self) -> dict[str, str]: ...

# Deprecated alias of QuerySet, for compatibility only.
_QuerySet: TypeAlias = QuerySet  # noqa: PYI047

class Prefetch(Generic[_LookupT, _PrefetchedQuerySetT, _ToAttrT]):
    prefetch_through: str
    prefetch_to: str
    queryset: _PrefetchedQuerySetT | None
    to_attr: _ToAttrT | None
    def __init__(
        self, lookup: _LookupT, queryset: _PrefetchedQuerySetT | None = None, to_attr: _ToAttrT | None = None
    ) -> None: ...
    if sys.version_info >= (3, 11):
        @override
        def __getstate__(self) -> dict[str, Any]: ...
    else:
        def __getstate__(self) -> dict[str, Any]: ...
    def add_prefix(self, prefix: str) -> None: ...
    def get_current_prefetch_to(self, level: int) -> str: ...
    def get_current_to_attr(self, level: int) -> tuple[str, str]: ...
    def get_current_querysets(self, level: int) -> list[_PrefetchedQuerySetT] | None: ...

def normalize_prefetch_lookups(lookups: Sequence[str | Prefetch], prefix: str | None = None) -> list[Prefetch]: ...
def prefetch_related_objects(model_instances: Sequence[_Model], *related_lookups: str | Prefetch) -> None: ...
async def aprefetch_related_objects(model_instances: Sequence[_Model], *related_lookups: str | Prefetch) -> None: ...
def get_prefetcher(instance: Model, through_attr: str, to_attr: str) -> tuple[Any, Any, bool, bool]: ...
def prefetch_one_level(
    instances: Sequence[Model], prefetcher: Any, lookup: Prefetch, level: int
) -> tuple[list[Any], list[Prefetch]]: ...

class InstanceCheckMeta(type): ...

class EmptyQuerySet(metaclass=InstanceCheckMeta):
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...

class RelatedPopulator:
    db: str
    cols_start: int
    cols_end: int
    init_list: list[str]
    reorder_for_init: Any
    model_cls: type[Model]
    pk_idx: int
    related_populators: list[RelatedPopulator]
    local_setter: Any
    remote_setter: Any
    def __init__(self, klass_info: dict[str, Any], select: Any, db: str) -> None: ...
    def populate(self, row: Sequence[Any], from_obj: Model) -> None: ...

def get_related_populators(klass_info: dict[str, Any], select: Any, db: str) -> list[RelatedPopulator]: ...



from collections.abc import Callable
from typing import Any

def get_callable(lookup_view: Callable[..., Any] | str) -> Callable[..., Any]: ...
def get_mod_func(callback: str) -> tuple[str, str]: ...



from collections.abc import Sequence
from typing import ClassVar

from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.operations.base import Operation
from django.db.migrations.state import ProjectState
from typing_extensions import Self

class Migration:
    operations: ClassVar[Sequence[Operation]]
    dependencies: ClassVar[Sequence[tuple[str, str]]]
    run_before: ClassVar[Sequence[tuple[str, str]]]
    replaces: ClassVar[Sequence[tuple[str, str]]]
    initial: ClassVar[bool | None]
    atomic: ClassVar[bool]
    name: str
    app_label: str
    def __init__(self, name: str, app_label: str) -> None: ...
    def mutate_state(self, project_state: ProjectState, preserve: bool = True) -> ProjectState: ...
    def apply(
        self, project_state: ProjectState, schema_editor: BaseDatabaseSchemaEditor, collect_sql: bool = False
    ) -> ProjectState: ...
    def unapply(
        self, project_state: ProjectState, schema_editor: BaseDatabaseSchemaEditor, collect_sql: bool = False
    ) -> ProjectState: ...
    def suggest_name(self) -> str: ...

class SwappableTuple(tuple[str, str]):
    setting: str
    def __new__(cls, value: tuple[str, str], setting: str) -> Self: ...

def swappable_dependency(value: str) -> SwappableTuple: ...



from django.core.exceptions import FieldDoesNotExist
from mypy.plugin import MethodContext
from mypy.types import AnyType, Instance, TypeOfAny, get_proper_type
from mypy.types import Type as MypyType

from mypy_django_plugin.django.context import DjangoContext, get_field_type_from_model_type_info
from mypy_django_plugin.lib import helpers
from mypy_django_plugin.lib.helpers import DjangoModel


def return_proper_field_type_from_get_field(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    if not (
        isinstance(ctx.type, Instance)
        and ctx.type.args
        and isinstance(model_type := get_proper_type(ctx.type.args[0]), Instance)
        and (field_name_expr := helpers.get_call_argument_by_name(ctx, "field_name")) is not None
        and (field_name := helpers.resolve_string_attribute_value(field_name_expr, django_context)) is not None
    ):
        return ctx.default_return_type

    field_type = get_field_type_from_model_type_info(model_type.type, field_name)
    if field_type is not None:
        return field_type

    if (django_model := DjangoModel.from_model_type(model_type, django_context)) is None:
        return ctx.default_return_type

    try:
        field = django_model.cls._meta.get_field(field_name)
        if field_info := helpers.lookup_class_typeinfo(helpers.get_typechecker_api(ctx), field.__class__):
            return Instance(field_info, [])
    except FieldDoesNotExist as e:
        ctx.api.fail(str(e), ctx.context)
        return AnyType(TypeOfAny.from_error)

    return ctx.default_return_type



from django.core.exceptions import ImproperlyConfigured as ImproperlyConfigured

class TaskException(Exception): ...
class InvalidTask(TaskException): ...
class InvalidTaskBackend(ImproperlyConfigured): ...
class TaskResultDoesNotExist(TaskException): ...
class TaskResultMismatch(TaskException): ...



import zoneinfo
from contextlib import ContextDecorator
from datetime import date, datetime, time, timedelta, timezone, tzinfo
from types import TracebackType
from typing import Literal, overload

def get_fixed_timezone(offset: timedelta | int) -> timezone: ...
def get_default_timezone() -> zoneinfo.ZoneInfo: ...
def get_default_timezone_name() -> str: ...
def get_current_timezone() -> zoneinfo.ZoneInfo: ...
def get_current_timezone_name() -> str: ...
def activate(timezone: tzinfo | str) -> None: ...
def deactivate() -> None: ...

class override(ContextDecorator):
    timezone: str | tzinfo | None
    old_timezone: tzinfo | None
    def __init__(self, timezone: str | tzinfo | None) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None: ...

def localtime(value: datetime | None = None, timezone: tzinfo | None = None) -> datetime: ...
def localdate(value: datetime | None = None, timezone: tzinfo | None = None) -> date: ...
def now() -> datetime: ...
@overload
def is_aware(value: time) -> Literal[False]: ...
@overload
def is_aware(value: datetime) -> bool: ...
@overload
def is_naive(value: time) -> Literal[True]: ...
@overload
def is_naive(value: datetime) -> bool: ...
def make_aware(value: datetime, timezone: tzinfo | None = None) -> datetime: ...
def make_naive(value: datetime, timezone: tzinfo | None = None) -> datetime: ...

__all__ = [
    "activate",
    "deactivate",
    "get_current_timezone",
    "get_current_timezone_name",
    "get_default_timezone",
    "get_default_timezone_name",
    "get_fixed_timezone",
    "is_aware",
    "is_naive",
    "localdate",
    "localtime",
    "make_aware",
    "make_naive",
    "now",
    "override",
]



from typing import Any

from django.tasks.base import Task, TaskResult
from typing_extensions import override

from .base import BaseTaskBackend

class DummyBackend(BaseTaskBackend):
    results: list[TaskResult]
    def __init__(self, alias: str, params: dict[str, Any]) -> None: ...
    @override
    def enqueue(self, task: Task, args: list[Any], kwargs: dict[str, Any]) -> TaskResult: ...
    @override
    def get_result(self, result_id: str) -> TaskResult: ...
    @override
    async def aget_result(self, result_id: str) -> TaskResult: ...
    def clear(self) -> None: ...



from datetime import datetime
from typing import Any, ClassVar, TypeVar

from django.contrib.sessions.backends.base import SessionBase
from django.db import models
from typing_extensions import Self

_T = TypeVar("_T", bound=AbstractBaseSession)

class BaseSessionManager(models.Manager[_T]):
    def encode(self, session_dict: dict[str, Any]) -> str: ...
    def save(self, session_key: str, session_dict: dict[str, Any], expire_date: datetime) -> _T: ...

class AbstractBaseSession(models.Model):
    session_key = models.CharField(primary_key=True)
    session_data = models.TextField()
    expire_date = models.DateTimeField()
    objects: ClassVar[BaseSessionManager[Self]]

    class Meta:
        abstract: ClassVar[bool]
        verbose_name: ClassVar[str]
        verbose_name_plural: ClassVar[str]

    @classmethod
    def get_session_store_class(cls) -> type[SessionBase] | None: ...
    def get_decoded(self) -> dict[str, Any]: ...



from collections.abc import Callable, Iterable
from typing import Any, Protocol, TypeAlias, TypeVar, type_check_only

from django.http.response import HttpResponseBase
from django.utils.deprecation import MiddlewareMixin
from django.views.generic.base import View

@type_check_only
class _MiddlewareClass(Protocol):
    def __call__(self, get_response: Callable[..., Any], *args: Any, **kwargs: Any) -> Any: ...

_ViewType = TypeVar("_ViewType", bound=View | Callable[..., Any])  # Any callable
_CallableType = TypeVar("_CallableType", bound=Callable[..., Any])
_ViewDecorator: TypeAlias = Callable[
    [Callable[..., HttpResponseBase]],
    Callable[..., HttpResponseBase],
]
_DecoratorFactory: TypeAlias = Callable[..., _ViewDecorator]

classonlymethod = classmethod

def method_decorator(
    decorator: _ViewDecorator | _DecoratorFactory | Iterable[_ViewDecorator | _DecoratorFactory],
    name: str = "",
) -> Callable[[_ViewType], _ViewType]: ...
def decorator_from_middleware_with_args(middleware_class: _MiddlewareClass) -> _DecoratorFactory: ...
def decorator_from_middleware(middleware_class: _MiddlewareClass) -> _ViewDecorator: ...
def make_middleware_decorator(middleware_class: type[MiddlewareMixin]) -> _DecoratorFactory: ...
def sync_and_async_middleware(func: _CallableType) -> _CallableType: ...
def sync_only_middleware(func: _CallableType) -> _CallableType: ...
def async_only_middleware(func: _CallableType) -> _CallableType: ...



from django.utils.connection import BaseConnectionHandler

from .backends.base import BaseTaskBackend
from .base import DEFAULT_TASK_BACKEND_ALIAS as DEFAULT_TASK_BACKEND_ALIAS
from .base import DEFAULT_TASK_QUEUE_NAME as DEFAULT_TASK_QUEUE_NAME
from .base import Task as Task
from .base import TaskContext as TaskContext
from .base import TaskResult as TaskResult
from .base import TaskResultStatus as TaskResultStatus
from .base import task as task

__all__ = [
    "DEFAULT_TASK_BACKEND_ALIAS",
    "DEFAULT_TASK_QUEUE_NAME",
    "Task",
    "TaskContext",
    "TaskResult",
    "TaskResultStatus",
    "default_task_backend",
    "task",
    "task_backends",
]

class TaskBackendHandler(BaseConnectionHandler[BaseTaskBackend]): ...

task_backends: TaskBackendHandler
# Actually ConnectionProxy, but quacks exactly like BaseTaskBackend, it's not worth distinguishing the two.
default_task_backend: BaseTaskBackend



from collections.abc import Iterator, Mapping, Sequence
from datetime import date as real_date
from typing import Any, NamedTuple

from django.http.request import QueryDict
from django.template.base import FilterExpression, Parser, PartialTemplate, Token
from django.template.context import Context
from django.utils.safestring import SafeString
from typing_extensions import override

from .base import Node, NodeList
from .library import Library
from .smartif import IfParser, Literal

register: Any

class AutoEscapeControlNode(Node):
    nodelist: NodeList
    setting: bool
    def __init__(self, setting: bool, nodelist: NodeList) -> None: ...

class CommentNode(Node): ...
class CsrfTokenNode(Node): ...

class CycleNode(Node):
    cyclevars: list[FilterExpression]
    variable_name: str | None
    silent: bool
    def __init__(
        self, cyclevars: list[FilterExpression], variable_name: str | None = None, silent: bool = False
    ) -> None: ...
    def reset(self, context: Context) -> None: ...

class DebugNode(Node): ...

class FilterNode(Node):
    filter_expr: FilterExpression
    nodelist: NodeList
    def __init__(self, filter_expr: FilterExpression, nodelist: NodeList) -> None: ...

class FirstOfNode(Node):
    vars: list[FilterExpression]
    asvar: str | None
    def __init__(self, variables: list[FilterExpression], asvar: str | None = None) -> None: ...

class ForNode(Node):
    loopvars: list[str] | str
    sequence: FilterExpression | str
    child_nodelists: Any
    is_reversed: bool
    nodelist_loop: list[str] | NodeList
    nodelist_empty: list[str] | NodeList
    def __init__(
        self,
        loopvars: list[str] | str,
        sequence: FilterExpression | str,
        is_reversed: bool,
        nodelist_loop: list[str] | NodeList,
        nodelist_empty: list[str] | NodeList | None = None,
    ) -> None: ...

class IfChangedNode(Node):
    nodelist_false: NodeList
    nodelist_true: NodeList
    child_nodelists: Any
    def __init__(self, nodelist_true: NodeList, nodelist_false: NodeList, *varlist: Any) -> None: ...

class IfNode(Node):
    conditions_nodelists: list[tuple[TemplateLiteral | None, NodeList]]
    def __init__(self, conditions_nodelists: list[tuple[TemplateLiteral | None, NodeList]]) -> None: ...
    def __iter__(self) -> Iterator[Node]: ...
    @property
    def nodelist(self) -> NodeList: ...

class LoremNode(Node):
    common: bool
    count: FilterExpression
    method: str
    def __init__(self, count: FilterExpression, method: str, common: bool) -> None: ...

class GroupedResult(NamedTuple):
    grouper: Any
    list: list[Any]

class RegroupNode(Node):
    expression: FilterExpression
    target: FilterExpression
    var_name: str
    def __init__(self, target: FilterExpression, expression: FilterExpression, var_name: str) -> None: ...
    def resolve_expression(self, obj: dict[str, real_date], context: Context) -> int | str: ...

class LoadNode(Node): ...

class NowNode(Node):
    format_string: str
    asvar: str | None
    def __init__(self, format_string: str, asvar: str | None = None) -> None: ...

class PartialDefNode(Node):
    partial_name: str
    inline: bool
    nodelist: NodeList
    def __init__(self, partial_name: str, inline: bool, nodelist: NodeList) -> None: ...

class PartialNode(Node):
    partial_name: str
    partial_mapping: Mapping[str, PartialTemplate]
    def __init__(self, partial_name: str, partial_mapping: Mapping[str, PartialTemplate]) -> None: ...

class ResetCycleNode(Node):
    node: CycleNode
    def __init__(self, node: CycleNode) -> None: ...

class SpacelessNode(Node):
    nodelist: NodeList
    def __init__(self, nodelist: NodeList) -> None: ...

class TemplateTagNode(Node):
    mapping: Any
    tagtype: str
    def __init__(self, tagtype: str) -> None: ...

class URLNode(Node):
    view_name: FilterExpression
    args: list[FilterExpression]
    kwargs: dict[str, FilterExpression]
    asvar: str | None
    def __init__(
        self,
        view_name: FilterExpression,
        args: list[FilterExpression],
        kwargs: dict[str, FilterExpression],
        asvar: str | None,
    ) -> None: ...

class VerbatimNode(Node):
    content: SafeString
    def __init__(self, content: SafeString) -> None: ...

class WidthRatioNode(Node):
    val_expr: FilterExpression
    max_expr: FilterExpression
    max_width: FilterExpression
    asvar: str | None
    def __init__(
        self,
        val_expr: FilterExpression,
        max_expr: FilterExpression,
        max_width: FilterExpression,
        asvar: str | None = None,
    ) -> None: ...

class WithNode(Node):
    nodelist: NodeList
    extra_context: dict[str, Any]
    def __init__(
        self,
        var: str | None,
        name: str | None,
        nodelist: NodeList | Sequence[Node],
        extra_context: dict[str, Any] | None = None,
    ) -> None: ...

def autoescape(parser: Parser, token: Token) -> AutoEscapeControlNode: ...
def comment(parser: Parser, token: Token) -> CommentNode: ...
def cycle(parser: Parser, token: Token) -> CycleNode: ...
def csrf_token(parser: Parser, token: Token) -> CsrfTokenNode: ...
def debug(parser: Parser, token: Token) -> DebugNode: ...
def do_filter(parser: Parser, token: Token) -> FilterNode: ...
def firstof(parser: Parser, token: Token) -> FirstOfNode: ...
def do_for(parser: Parser, token: Token) -> ForNode: ...

class TemplateLiteral(Literal):
    text: str
    def __init__(self, value: FilterExpression, text: str) -> None: ...
    @override
    def display(self) -> str: ...

class TemplateIfParser(IfParser):
    current_token: TemplateLiteral
    pos: int
    tokens: list[TemplateLiteral]
    error_class: Any
    template_parser: Parser
    def __init__(self, parser: Parser, *args: Any, **kwargs: Any) -> None: ...

def do_if(parser: Parser, token: Token) -> IfNode: ...
def ifchanged(parser: Parser, token: Token) -> IfChangedNode: ...
def find_library(parser: Parser, name: str) -> Library: ...
def load_from_library(library: Library, label: str, names: list[str]) -> Library: ...
def load(parser: Parser, token: Token) -> LoadNode: ...
def lorem(parser: Parser, token: Token) -> LoremNode: ...
def now(parser: Parser, token: Token) -> NowNode: ...
def partial_func(parser: Parser, token: Token) -> PartialNode: ...
def partialdef_func(parser: Parser, token: Token) -> PartialDefNode: ...
def querystring(context: Context, query_dict: QueryDict | None = None, **kwargs: Any) -> str: ...
def regroup(parser: Parser, token: Token) -> RegroupNode: ...
def resetcycle(parser: Parser, token: Token) -> ResetCycleNode: ...
def spaceless(parser: Parser, token: Token) -> SpacelessNode: ...
def templatetag(parser: Parser, token: Token) -> TemplateTagNode: ...
def url(parser: Parser, token: Token) -> URLNode: ...
def verbatim(parser: Parser, token: Token) -> VerbatimNode: ...
def widthratio(parser: Parser, token: Token) -> WidthRatioNode: ...
def do_with(parser: Parser, token: Token) -> WithNode: ...



from collections.abc import Sequence
from typing import Any, ClassVar

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models.expressions import Func
from django.db.models.fields.json import JSONField
from django.db.models.sql.compiler import SQLCompiler, _AsSqlType
from typing_extensions import override

class JSONArray(Func):
    output_field: ClassVar[JSONField]
    @override
    def as_sql(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any) -> _AsSqlType: ...  # type: ignore [override]
    def as_native(
        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, *, returning: str, **extra_context: Any
    ) -> _AsSqlType: ...
    def as_postgresql(
        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any
    ) -> _AsSqlType: ...
    def as_oracle(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any) -> _AsSqlType: ...

class JSONObject(Func):
    output_field: ClassVar[JSONField]
    def __init__(self, **fields: Any) -> None: ...
    @override
    def as_sql(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any) -> _AsSqlType: ...  # type: ignore [override]
    def as_native(
        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, *, returning: str, **extra_context: Any
    ) -> _AsSqlType: ...
    def as_postgresql(
        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any
    ) -> _AsSqlType: ...
    def as_oracle(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any) -> _AsSqlType: ...
    def join(self, args: Sequence[Any]) -> str: ...



from collections.abc import Iterator
from contextlib import AbstractContextManager
from typing import Any

from django.db.backends.base.base import BaseDatabaseWrapper
from django.utils.functional import cached_property
from typing_extensions import override

from .client import DatabaseClient
from .creation import DatabaseCreation
from .features import DatabaseFeatures
from .introspection import DatabaseIntrospection
from .operations import DatabaseOperations
from .validation import DatabaseValidation

def wrap_oracle_errors() -> AbstractContextManager[None]: ...

class _UninitializedOperatorsDescriptor:
    def __get__(self, instance: Any, cls: Any | None = None) -> Any: ...

class DatabaseWrapper(BaseDatabaseWrapper):
    client: DatabaseClient
    creation: DatabaseCreation
    features: DatabaseFeatures
    introspection: DatabaseIntrospection
    validation: DatabaseValidation

    client_class: type[DatabaseClient]
    creation_class: type[DatabaseCreation]
    features_class: type[DatabaseFeatures]
    introspection_class: type[DatabaseIntrospection]
    ops_class: type[DatabaseOperations]
    validation_class: type[DatabaseValidation]

    vendor: str
    display_name: str
    data_types: Any
    data_type_check_constraints: Any
    operators: Any
    pattern_esc: str
    Database: Any
    SchemaEditorClass: Any
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    @property
    def is_pool(self) -> bool: ...
    @property
    def pool(self) -> Any: ...
    def close_pool(self) -> None: ...
    @override
    def get_connection_params(self) -> Any: ...
    @override
    def get_new_connection(self, conn_params: Any) -> Any: ...
    pattern_ops: Any
    @override
    def init_connection_state(self) -> None: ...
    @override
    def create_cursor(self, name: Any | None = None) -> Any: ...
    @override
    def check_constraints(self, table_names: Any | None = None) -> None: ...
    @override
    def is_usable(self) -> bool: ...
    @cached_property
    def oracle_version(self) -> tuple[int, ...]: ...
    @cached_property
    def oracledb_version(self) -> tuple[int, ...]: ...

class OracleParam:
    force_bytes: Any
    input_size: Any
    def __init__(self, param: Any, cursor: Any, strings_only: bool = False) -> None: ...

class VariableWrapper:
    var: Any
    def __init__(self, var: Any) -> None: ...
    def bind_parameter(self, cursor: Any) -> Any: ...
    def __getattr__(self, key: Any) -> Any: ...
    @override
    def __setattr__(self, key: Any, value: Any) -> None: ...

class FormatStylePlaceholderCursor:
    charset: str
    cursor: Any
    database: BaseDatabaseWrapper
    def __init__(self, connection: Any, database: BaseDatabaseWrapper) -> None: ...
    def execute(self, query: Any, params: Any | None = None) -> Any: ...
    def executemany(self, query: Any, params: Any | None = None) -> Any: ...
    def close(self) -> None: ...
    def var(self, *args: Any) -> Any: ...
    def arrayvar(self, *args: Any) -> Any: ...
    def __getattr__(self, attr: Any) -> Any: ...
    def __iter__(self) -> Iterator[Any]: ...



from typing import Any, ClassVar

from django.contrib import admin

class RedirectAdmin(admin.ModelAdmin):
    list_display: Any
    list_filter: ClassVar[Any]
    search_fields: ClassVar[Any]
    radio_fields: ClassVar[Any]



from typing import Any

from django.core.management import BaseCommand
from typing_extensions import override

class Command(BaseCommand):
    shells: list[str]

    @override
    def handle(self, **options: Any) -> None: ...
    def ipython(self, options: Any) -> None: ...
    def bpython(self, options: Any) -> None: ...
    def python(self, options: Any) -> None: ...
    def get_auto_imports(self) -> list[str] | None: ...
    def get_namespace(self, **options: Any) -> dict[str, Any]: ...



from collections.abc import Sequence
from typing import Any

from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.operations.base import Operation
from django.db.models.base import Model
from django.db.models.constraints import BaseConstraint
from django.db.models.fields import Field
from django.db.models.indexes import Index
from django.db.models.manager import Manager
from django.db.models.options import _OptionTogetherT
from django.utils.functional import cached_property
from typing_extensions import override

class ModelOperation(Operation):
    name: str
    def __init__(self, name: str) -> None: ...
    @cached_property
    def name_lower(self) -> str: ...
    def can_reduce_through(self, operation: Operation, app_label: str) -> bool: ...

class CreateModel(ModelOperation):
    fields: list[tuple[str, Field]]
    options: dict[str, Any]
    bases: Sequence[type[Model] | str] | None
    managers: Sequence[tuple[str, Manager]] | None
    def __init__(
        self,
        name: str,
        fields: list[tuple[str, Field]],
        options: dict[str, Any] | None = None,
        bases: Sequence[type[Any] | str] | None = None,
        managers: Sequence[tuple[str, Manager]] | None = None,
    ) -> None: ...

class DeleteModel(ModelOperation): ...

class RenameModel(ModelOperation):
    old_name: str
    new_name: str
    def __init__(self, old_name: str, new_name: str) -> None: ...
    @cached_property
    def old_name_lower(self) -> str: ...
    @cached_property
    def new_name_lower(self) -> str: ...

class ModelOptionOperation(ModelOperation): ...

class AlterModelTable(ModelOptionOperation):
    table: str | None
    def __init__(self, name: str, table: str | None) -> None: ...

class AlterModelTableComment(ModelOptionOperation):
    table_comment: str
    def __init__(self, name: str, table_comment: str) -> None: ...
    @override
    def state_forwards(self, app_label: str, state: Any) -> None: ...
    @override
    def database_forwards(
        self, app_label: str, schema_editor: BaseDatabaseSchemaEditor, from_state: Any, to_state: Any
    ) -> None: ...
    @override
    def database_backwards(
        self, app_label: str, schema_editor: BaseDatabaseSchemaEditor, from_state: Any, to_state: Any
    ) -> None: ...
    @override
    def describe(self) -> str: ...
    @property
    @override
    def migration_name_fragment(self) -> str: ...

class AlterTogetherOptionOperation(ModelOptionOperation):
    option_name: str
    def __init__(
        self,
        name: str,
        option_value: _OptionTogetherT | None,
    ) -> None: ...
    @cached_property
    def option_value(self) -> set[tuple[str, ...]] | None: ...
    @override
    def state_forwards(self, app_label: str, state: Any) -> None: ...
    @override
    def database_forwards(
        self, app_label: str, schema_editor: BaseDatabaseSchemaEditor, from_state: Any, to_state: Any
    ) -> None: ...
    @override
    def database_backwards(
        self, app_label: str, schema_editor: BaseDatabaseSchemaEditor, from_state: Any, to_state: Any
    ) -> None: ...
    @override
    def references_field(self, model_name: str, name: str, app_label: str) -> bool: ...
    @override
    def describe(self) -> str: ...
    @property
    @override
    def migration_name_fragment(self) -> str: ...

class AlterUniqueTogether(AlterTogetherOptionOperation):
    option_name: str
    unique_together: set[tuple[str, ...]] | None
    def __init__(self, name: str, unique_together: _OptionTogetherT | None) -> None: ...

class AlterIndexTogether(AlterTogetherOptionOperation):
    option_name: str
    index_together: set[tuple[str, ...]] | None
    def __init__(self, name: str, index_together: _OptionTogetherT | None) -> None: ...

class AlterOrderWithRespectTo(ModelOptionOperation):
    option_name: str
    order_with_respect_to: str | None
    def __init__(self, name: str, order_with_respect_to: str | None) -> None: ...

class AlterModelOptions(ModelOptionOperation):
    ALTER_OPTION_KEYS: list[str]
    options: dict[str, Any]
    def __init__(self, name: str, options: dict[str, Any]) -> None: ...

class AlterModelManagers(ModelOptionOperation):
    managers: Sequence[tuple[str, Manager]]
    def __init__(self, name: str, managers: Sequence[tuple[str, Manager]]) -> None: ...

class IndexOperation(Operation):
    option_name: str
    @cached_property
    def model_name_lower(self) -> str: ...

class AddIndex(IndexOperation):
    model_name: str
    index: Index
    def __init__(self, model_name: str, index: Index) -> None: ...

class RemoveIndex(IndexOperation):
    model_name: str
    name: str
    def __init__(self, model_name: str, name: str) -> None: ...

class RenameIndex(IndexOperation):
    model_name: str
    new_name: str
    old_name: str | None
    old_fields: Sequence[str] | None
    def __init__(
        self,
        model_name: str,
        new_name: str,
        old_name: str | None = None,
        old_fields: Sequence[str] | None = None,
    ) -> None: ...
    @cached_property
    def old_name_lower(self) -> str: ...
    @cached_property
    def new_name_lower(self) -> str: ...

class AddConstraint(IndexOperation):
    model_name: str
    constraint: BaseConstraint
    def __init__(self, model_name: str, constraint: BaseConstraint) -> None: ...

class RemoveConstraint(IndexOperation):
    model_name: str
    name: str
    def __init__(self, model_name: str, name: str) -> None: ...

class AlterConstraint(IndexOperation):
    model_name: str
    name: str
    constraint: BaseConstraint
    def __init__(self, model_name: str, name: str, constraint: BaseConstraint) -> None: ...



from typing import Any

from django.utils.connection import BaseConnectionHandler
from typing_extensions import override

from .backends.base import BaseCache as BaseCache
from .backends.base import CacheKeyWarning as CacheKeyWarning
from .backends.base import InvalidCacheBackendError as InvalidCacheBackendError
from .backends.base import InvalidCacheKey as InvalidCacheKey

DEFAULT_CACHE_ALIAS: str

class CacheHandler(BaseConnectionHandler[BaseCache]):
    settings_name: str
    exception_class: type[Exception]
    @override
    def create_connection(self, alias: str) -> BaseCache: ...
    @override
    def all(self, initialized_only: bool = False) -> list[BaseCache]: ...

def close_caches(**kwargs: Any) -> None: ...

caches: CacheHandler
# Actually ConnectionProxy, but quacks exactly like BaseCache, it's not worth distinguishing the two.
cache: BaseCache

__all__ = [
    "DEFAULT_CACHE_ALIAS",
    "BaseCache",
    "CacheKeyWarning",
    "InvalidCacheBackendError",
    "InvalidCacheKey",
    "cache",
    "caches",
]



import threading
from collections.abc import Callable, Iterable
from typing import Any

from django.db.models.base import Model

from .config import AppConfig

class Apps:
    all_models: dict[str, dict[str, type[Model]]]
    app_configs: dict[str, AppConfig]
    stored_app_configs: list[Any]
    apps_ready: bool
    ready_event: threading.Event
    loading: bool
    _pending_operations: dict[tuple[str, str], list]
    models_ready: bool
    ready: bool
    def __init__(self, installed_apps: Iterable[AppConfig | str] | None = ()) -> None: ...
    def populate(self, installed_apps: Iterable[AppConfig | str] | None = None) -> None: ...
    def check_apps_ready(self) -> None: ...
    def check_models_ready(self) -> None: ...
    def get_app_configs(self) -> Iterable[AppConfig]: ...
    def get_app_config(self, app_label: str) -> AppConfig: ...
    # it's not possible to support it in plugin properly now
    def get_models(self, include_auto_created: bool = False, include_swapped: bool = False) -> list[type[Model]]: ...
    def get_model(self, app_label: str, model_name: str | None = None, require_ready: bool = True) -> type[Any]: ...
    def register_model(self, app_label: str, model: type[Model]) -> None: ...
    def is_installed(self, app_name: str) -> bool: ...
    def get_containing_app_config(self, object_name: str) -> AppConfig | None: ...
    def get_registered_model(self, app_label: str, model_name: str) -> type[Model]: ...
    def get_swappable_settings_name(self, to_string: str) -> str | None: ...
    def set_available_apps(self, available: Iterable[str]) -> None: ...
    def unset_available_apps(self) -> None: ...
    def set_installed_apps(self, installed: Iterable[str]) -> None: ...
    def unset_installed_apps(self) -> None: ...
    def clear_cache(self) -> None: ...
    def lazy_model_operation(self, function: Callable, *model_keys: Any) -> None: ...
    def do_pending_operations(self, model: type[Model]) -> None: ...

apps: Apps



from typing import NamedTuple

from mypy.nodes import AssignmentStmt, NameExpr, Node, TypeInfo
from mypy.plugin import FunctionContext, MethodContext
from mypy.types import Instance, ProperType, UninhabitedType, get_proper_type
from mypy.types import Type as MypyType

from mypy_django_plugin.django.context import DjangoContext
from mypy_django_plugin.lib import fullnames, helpers


class M2MThrough(NamedTuple):
    arg: Node | None
    model: ProperType


class M2MTo(NamedTuple):
    arg: Node
    model: ProperType
    self: bool  # ManyToManyField('self', ...)


class M2MArguments(NamedTuple):
    to: M2MTo
    through: M2MThrough | None


def fill_model_args_for_many_to_many_field(
    *,
    ctx: FunctionContext,
    model_info: TypeInfo,
    django_context: DjangoContext,
) -> MypyType:
    default_return_type = get_proper_type(ctx.default_return_type)
    if not (
        ctx.args
        and ctx.args[0]
        and isinstance(default_return_type, Instance)
        and len(default_return_type.args) >= 2
        and (args := get_m2m_arguments(ctx=ctx, model_info=model_info, django_context=django_context))
    ):
        return ctx.default_return_type

    default_to_arg = get_proper_type(default_return_type.args[0])
    to_arg: MypyType
    if isinstance(default_to_arg, UninhabitedType):
        to_arg = args.to.model
    else:
        # Avoid overwriting a decent 'to' argument
        to_arg = default_return_type.args[0]

    default_through_arg = get_proper_type(default_return_type.args[1])
    if isinstance(default_through_arg, UninhabitedType):
        if helpers.is_abstract_model(model_info):
            # Many to many on abstract models doesn't create any implicit, concrete
            # through model, so we populate it with the upper bound to avoid error messages
            through_arg = default_return_type.type.defn.type_vars[1].upper_bound
        elif args.through is None:
            through_arg = default_through_arg
        else:
            through_arg = args.through.model
    else:
        # Avoid overwriting a decent 'through' argument
        through_arg = default_return_type.args[1]

    return default_return_type.copy_modified(args=[to_arg, through_arg])


def get_m2m_arguments(
    *,
    ctx: FunctionContext,
    model_info: TypeInfo,
    django_context: DjangoContext,
) -> M2MArguments | None:
    checker = helpers.get_typechecker_api(ctx)
    to_arg = ctx.args[0][0]
    to_model = helpers.get_model_from_expression(
        to_arg, self_model=model_info, api=checker, django_context=django_context
    )
    if to_model is None:
        # 'ManyToManyField()' requires the 'to' argument
        return None
    to = M2MTo(arg=to_arg, model=to_model, self=to_model.type == model_info)

    through = None
    if len(ctx.args) > 5 and ctx.args[5]:
        # 'ManyToManyField(..., through=)' was called
        through_arg = ctx.args[5][0]
        through_model = helpers.get_model_from_expression(
            through_arg, self_model=model_info, api=checker, django_context=django_context
        )
        if through_model is not None:
            through = M2MThrough(arg=through_arg, model=through_model)
    elif not helpers.is_abstract_model(model_info):
        # No explicit 'through' argument was provided and model is concrete. We need
        # to dig up any generated through model for this 'ManyToManyField()' field
        through_arg = None
        m2m_throughs = helpers.get_django_metadata(model_info).get("m2m_throughs", {})
        if m2m_throughs:
            field_name = None
            for defn in model_info.defn.defs.body:
                if (
                    isinstance(defn, AssignmentStmt)
                    and defn.rvalue is ctx.context
                    and len(defn.lvalues) == 1
                    and isinstance(defn.lvalues[0], NameExpr)
                ):
                    field_name = defn.lvalues[0].name
                    break

            if field_name is not None:
                through_model_fullname = m2m_throughs.get(field_name)
                if through_model_fullname is not None:
                    through_model_info = helpers.lookup_fully_qualified_typeinfo(checker, through_model_fullname)
                    if through_model_info is not None:
                        through = M2MThrough(arg=through_arg, model=Instance(through_model_info, []))

    return M2MArguments(to=to, through=through)


def get_related_manager_and_model(ctx: MethodContext) -> tuple[Instance, Instance, Instance] | None:
    """
    Returns a 3-tuple consisting of:
      1. A `ManyRelatedManager` instance
      2. The first type parameter (_To) instance of 1. when it's a model
      3. The second type parameter (_Through) instance of 1. when it's a model
    When encountering a `ManyRelatedManager` that has populated its 2 first type
    parameters with models. Otherwise `None` is returned.

    For example: if given a `ManyRelatedManager[A, B]` where `A` and `B` are models the
    following 3-tuple is returned: `(ManyRelatedManager[A, B], A, B)`.
    """
    default_return_type = get_proper_type(ctx.default_return_type)
    if (
        isinstance(default_return_type, Instance)
        and default_return_type.type.fullname == fullnames.MANY_RELATED_MANAGER
    ):
        # This is a call to '__get__' overload with a model instance of 'ManyToManyDescriptor'.
        # Returning a 'ManyRelatedManager'. Which we want to, just like Django, build from the
        # default manager of the related model.
        many_related_manager = default_return_type
        # Require first and second type argument of 'ManyRelatedManager' to be models
        if (
            len(many_related_manager.args) >= 2
            and isinstance((_To := get_proper_type(many_related_manager.args[0])), Instance)
            and helpers.is_model_type(_To.type)
            and isinstance((_Through := get_proper_type(many_related_manager.args[1])), Instance)
            and helpers.is_model_type(_Through.type)
        ):
            return many_related_manager, _To, _Through

    return None


def refine_many_to_many_related_manager(ctx: MethodContext) -> MypyType:
    """
    Updates the 'ManyRelatedManager' returned by e.g. 'ManyToManyDescriptor' to be a subclass
    of 'ManyRelatedManager' and the related model's default manager.
    """
    related_objects = get_related_manager_and_model(ctx)
    if related_objects is None:
        return ctx.default_return_type

    _, related_model_instance, through_model_instance = related_objects
    checker = helpers.get_typechecker_api(ctx)
    related_manager_info = helpers.get_many_to_many_manager_info(
        checker, to=related_model_instance.type, derived_from="_default_manager"
    )
    if related_manager_info is None:
        return ctx.default_return_type
    return Instance(related_manager_info, [through_model_instance])



from typing import Any

from django.core.management.commands.inspectdb import Command as InspectDBCommand
from django.db.backends.base.base import BaseDatabaseWrapper
from typing_extensions import override

class Command(InspectDBCommand):
    @override
    def get_field_type(
        self, connection: BaseDatabaseWrapper, table_name: str, row: Any
    ) -> tuple[str, dict[str, Any], list[str]]: ...



from collections.abc import Iterable, Sequence
from typing import Any, Generic, Literal, TypeVar

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models.expressions import Combinable, Expression, Func
from django.db.models.fields import BooleanField
from django.db.models.query_utils import RegisterLookupMixin
from django.db.models.sql.compiler import SQLCompiler, _AsSqlType, _ParamT
from django.utils.datastructures import OrderedSet
from django.utils.functional import cached_property
from typing_extensions import override

_T = TypeVar("_T")

class Lookup(Expression, Generic[_T]):
    lookup_name: str | None
    prepare_rhs: bool
    can_use_none_as_rhs: bool
    lhs: Any
    rhs: Any
    bilateral_transforms: list[type[Transform]]
    def __init__(self, lhs: Any, rhs: Any) -> None: ...
    def apply_bilateral_transforms(self, value: Expression) -> Expression: ...
    def batch_process_rhs(
        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, rhs: OrderedSet | None = None
    ) -> tuple[list[str], list[str]]: ...
    @override
    def get_source_expressions(self) -> list[Expression]: ...
    @override
    def set_source_expressions(self, new_exprs: Sequence[Combinable | Expression]) -> None: ...
    def get_prep_lookup(self) -> Any: ...
    def get_prep_lhs(self) -> Any: ...
    def get_db_prep_lookup(self, value: _ParamT, connection: BaseDatabaseWrapper) -> _AsSqlType: ...
    def process_lhs(
        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, lhs: Expression | None = None
    ) -> _AsSqlType: ...
    def process_rhs(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...
    def rhs_is_direct_value(self) -> bool: ...
    def as_oracle(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...
    @cached_property
    @override
    def output_field(self) -> BooleanField: ...
    @property
    @override
    def identity(self) -> tuple[type[Lookup], Any, Any]: ...
    @cached_property
    @override
    def allowed_default(self) -> bool: ...  # type: ignore[override]

class Transform(RegisterLookupMixin, Func):
    bilateral: bool
    @property
    def lhs(self) -> Expression: ...
    def get_bilateral_transforms(self) -> list[type[Transform]]: ...

class BuiltinLookup(Lookup[_T]):
    @override
    def process_lhs(
        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, lhs: Expression | None = None
    ) -> _AsSqlType: ...
    def get_rhs_op(self, connection: BaseDatabaseWrapper, rhs: str) -> str: ...

class FieldGetDbPrepValueMixin:
    get_db_prep_lookup_value_is_iterable: bool
    def get_db_prep_lookup(self, value: _ParamT, connection: BaseDatabaseWrapper) -> _AsSqlType: ...

class FieldGetDbPrepValueIterableMixin(FieldGetDbPrepValueMixin):
    get_db_prep_lookup_value_is_iterable: Literal[True]
    def get_prep_lookup(self) -> Iterable[Any]: ...
    def process_rhs(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...
    def resolve_expression_parameter(
        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, sql: str, param: Any
    ) -> _AsSqlType: ...
    def batch_process_rhs(
        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, rhs: OrderedSet | None = None
    ) -> tuple[list[str], list[str]]: ...

class PostgresOperatorLookup(Lookup[_T]):
    postgres_operator: str | None
    def as_postgresql(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...

class Exact(FieldGetDbPrepValueMixin, BuiltinLookup[_T]): ...

class IExact(BuiltinLookup[_T]):
    @override
    def process_rhs(self, qn: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...

class GreaterThan(FieldGetDbPrepValueMixin, BuiltinLookup[_T]): ...
class GreaterThanOrEqual(FieldGetDbPrepValueMixin, BuiltinLookup[_T]): ...
class LessThan(FieldGetDbPrepValueMixin, BuiltinLookup[_T]): ...
class LessThanOrEqual(FieldGetDbPrepValueMixin, BuiltinLookup[_T]): ...

class IntegerFieldOverflow:
    underflow_exception: type[Exception]
    overflow_exception: type[Exception]
    def process_rhs(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...

class IntegerFieldFloatRounding:
    rhs: Any
    def get_prep_lookup(self) -> Any: ...

class IntegerFieldExact(IntegerFieldOverflow, Exact[int | float]): ...
class IntegerGreaterThan(IntegerFieldOverflow, GreaterThan[int | float]): ...
class IntegerGreaterThanOrEqual(IntegerFieldOverflow, IntegerFieldFloatRounding, GreaterThanOrEqual[int | float]): ...
class IntegerLessThan(IntegerFieldOverflow, IntegerFieldFloatRounding, LessThan[int | float]): ...
class IntegerLessThanOrEqual(IntegerFieldOverflow, LessThanOrEqual[int | float]): ...

class In(FieldGetDbPrepValueIterableMixin, BuiltinLookup[_T]):
    def split_parameter_list_as_sql(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> Any: ...

class PatternLookup(BuiltinLookup[str]):
    param_pattern: str
    @override
    def process_rhs(self, qn: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...

class Contains(PatternLookup): ...
class IContains(Contains): ...
class StartsWith(PatternLookup): ...
class IStartsWith(StartsWith): ...
class EndsWith(PatternLookup): ...
class IEndsWith(EndsWith): ...
class Range(FieldGetDbPrepValueIterableMixin, BuiltinLookup[_T]): ...
class IsNull(BuiltinLookup[bool]): ...
class Regex(BuiltinLookup[str]): ...
class IRegex(Regex): ...

class YearLookup(Lookup):
    def year_lookup_bounds(self, connection: BaseDatabaseWrapper, year: int) -> list[str]: ...
    def get_direct_rhs_sql(self, connection: BaseDatabaseWrapper, rhs: str) -> str: ...
    def get_bound_params(self, start: Any, finish: Any) -> Any: ...

class YearExact(YearLookup, Exact[_T]):
    @override
    def get_bound_params(self, start: Any, finish: Any) -> tuple[Any, Any]: ...

class YearGt(YearLookup, GreaterThan[_T]):
    @override
    def get_bound_params(self, start: Any, finish: Any) -> tuple[Any]: ...

class YearGte(YearLookup, GreaterThanOrEqual[_T]):
    @override
    def get_bound_params(self, start: Any, finish: Any) -> tuple[Any]: ...

class YearLt(YearLookup, LessThan[_T]):
    @override
    def get_bound_params(self, start: Any, finish: Any) -> tuple[Any]: ...

class YearLte(YearLookup, LessThanOrEqual[_T]):
    @override
    def get_bound_params(self, start: Any, finish: Any) -> tuple[Any]: ...

class UUIDTextMixin:
    rhs: Any
    def process_rhs(self, qn: Any, connection: BaseDatabaseWrapper) -> Any: ...

class UUIDIExact(UUIDTextMixin, IExact[_T]): ...
class UUIDContains(UUIDTextMixin, Contains): ...
class UUIDIContains(UUIDTextMixin, IContains): ...
class UUIDStartsWith(UUIDTextMixin, StartsWith): ...
class UUIDIStartsWith(UUIDTextMixin, IStartsWith): ...
class UUIDEndsWith(UUIDTextMixin, EndsWith): ...
class UUIDIEndsWith(UUIDTextMixin, IEndsWith): ...



from logging import Logger
from typing import Any

from django.dispatch import Signal as Signal
from django.tasks.backends.base import BaseTaskBackend
from django.tasks.base import TaskResult

from .base import TaskResultStatus as TaskResultStatus

logger: Logger

task_enqueued: Signal
task_finished: Signal
task_started: Signal

def clear_tasks_handlers(*, setting: str, **kwargs: Any) -> None: ...
def log_task_enqueued(sender: type[BaseTaskBackend], task_result: TaskResult, **kwargs: Any) -> None: ...
def log_task_started(sender: type[BaseTaskBackend], task_result: TaskResult, **kwargs: Any) -> None: ...
def log_task_finished(sender: type[BaseTaskBackend], task_result: TaskResult, **kwargs: Any) -> None: ...



from collections.abc import Awaitable, Callable, Iterable
from typing import TypeVar, overload

from django.contrib.auth.models import _AnyUser
from django.http.response import HttpResponseBase
from django.utils.functional import _StrOrPromise

_VIEW = TypeVar("_VIEW", bound=Callable[..., HttpResponseBase | Awaitable[HttpResponseBase]])

def user_passes_test(
    test_func: Callable[[_AnyUser], bool],
    login_url: _StrOrPromise | None = ...,
    redirect_field_name: str | None = ...,
) -> Callable[[_VIEW], _VIEW]: ...
@overload
def login_required(
    function: _VIEW,
    redirect_field_name: str | None = ...,
    login_url: _StrOrPromise | None = ...,
) -> _VIEW: ...
@overload
def login_required(
    function: None = None,
    redirect_field_name: str | None = ...,
    login_url: _StrOrPromise | None = ...,
) -> Callable[[_VIEW], _VIEW]: ...
def login_not_required(view_func: _VIEW) -> _VIEW: ...
def permission_required(
    perm: Iterable[str] | str, login_url: _StrOrPromise | None = ..., raise_exception: bool = ...
) -> Callable[[_VIEW], _VIEW]: ...



from typing import IO, Any

from django.core.serializers.base import Deserializer as PythonDeserializer
from django.core.serializers.python import Serializer as PythonSerializer
from django.db.models.fields import Field
from typing_extensions import override
from yaml import CSafeDumper as SafeDumper
from yaml import MappingNode, ScalarNode

class DjangoSafeDumper(SafeDumper):
    def represent_decimal(self, data: Any) -> ScalarNode: ...
    def represent_ordered_dict(self, data: Any) -> MappingNode: ...

class Serializer(PythonSerializer):
    internal_use_only: bool
    @override
    def handle_field(self, obj: Any, field: Field) -> None: ...
    @override
    def end_serialization(self) -> None: ...
    @override
    def getvalue(self) -> Any: ...

class Deserializer(PythonDeserializer):
    def __init__(self, stream_or_string: bytes | str | IO[bytes] | IO[str], **options: Any) -> None: ...



from collections.abc import Sequence
from typing import Any

from django.apps.config import AppConfig
from django.core.checks.messages import CheckMessage

def check_user_model(app_configs: Sequence[AppConfig] | None, **kwargs: Any) -> list[CheckMessage]: ...
def check_models_permissions(app_configs: Sequence[AppConfig] | None, **kwargs: Any) -> list[CheckMessage]: ...
def check_middleware(app_configs: Sequence[AppConfig] | None, **kwargs: Any) -> list[CheckMessage]: ...



from typing import Any

from django.core.management.base import BaseCommand
from typing_extensions import override

class Command(BaseCommand):
    output_transaction: bool

    @override
    def handle(self, **options: Any) -> str: ...



from typing import Any

from django.core.serializers.json import _DeserializerInput
from django.core.serializers.python import Deserializer as PythonDeserializer
from django.core.serializers.python import Serializer as PythonSerializer

class Serializer(PythonSerializer):
    json_kwargs: dict[str, Any]

class Deserializer(PythonDeserializer):
    def __init__(self, stream_or_string: _DeserializerInput, **options: Any) -> None: ...



from django.db.migrations.serializer import BaseSerializer
from typing_extensions import override

class RangeSerializer(BaseSerializer):
    @override
    def serialize(self) -> tuple[str, set[str]]: ...



from collections.abc import Sequence
from enum import Enum
from typing import Any, cast, overload

from django.core.checks import CheckMessage
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.models.base import Model
from django.db.models.expressions import BaseExpression, Combinable
from django.db.models.query_utils import Q
from django.utils.functional import _StrOrPromise
from typing_extensions import Self, override

class Deferrable(Enum):
    DEFERRED = cast(str, ...)
    IMMEDIATE = cast(str, ...)

class BaseConstraint:
    name: str
    violation_error_code: str | None
    violation_error_message: _StrOrPromise | None
    default_violation_error_message: _StrOrPromise
    non_db_attrs: tuple[str, ...]
    def __init__(
        self,
        *,
        name: str,
        violation_error_code: str | None = None,
        violation_error_message: _StrOrPromise | None = None,
    ) -> None: ...
    @property
    def contains_expressions(self) -> bool: ...
    def constraint_sql(self, model: type[Model] | None, schema_editor: BaseDatabaseSchemaEditor | None) -> str: ...
    def create_sql(self, model: type[Model] | None, schema_editor: BaseDatabaseSchemaEditor | None) -> str: ...
    def remove_sql(self, model: type[Model] | None, schema_editor: BaseDatabaseSchemaEditor | None) -> str: ...
    def validate(
        self, model: type[Model], instance: Model, exclude: set[str] | None = None, using: str = "default"
    ) -> None: ...
    def get_violation_error_message(self) -> _StrOrPromise | dict[str, Any]: ...
    def check(self, model: type[Model], connection: BaseDatabaseWrapper) -> list[CheckMessage]: ...
    def deconstruct(self) -> tuple[str, Sequence[Any], dict[str, Any]]: ...
    def clone(self) -> Self: ...

class CheckConstraint(BaseConstraint):
    condition: Q | BaseExpression

    def __init__(
        self,
        *,
        name: str,
        condition: Q | BaseExpression,
        violation_error_code: str | None = None,
        violation_error_message: _StrOrPromise | None = None,
    ) -> None: ...
    @override
    def check(self, model: type[Model], connection: BaseDatabaseWrapper) -> list[CheckMessage]: ...
    @override
    def validate(
        self, model: type[Model], instance: Model, exclude: set[str] | None = None, using: str = "default"
    ) -> None: ...

class UniqueConstraint(BaseConstraint):
    expressions: Sequence[BaseExpression | Combinable]
    fields: Sequence[str]
    condition: Q | None
    deferrable: Deferrable | None
    nulls_distinct: bool | None

    @overload
    def __init__(
        self,
        *expressions: str | BaseExpression | Combinable,
        fields: None = None,
        name: str | None = None,
        condition: Q | None = None,
        deferrable: Deferrable | None = None,
        include: Sequence[str] | None = None,
        opclasses: Sequence[Any] = (),
        nulls_distinct: bool | None = None,
        violation_error_code: str | None = None,
        violation_error_message: _StrOrPromise | None = None,
    ) -> None: ...
    @overload
    def __init__(
        self,
        *,
        fields: Sequence[str],
        name: str | None = None,
        condition: Q | None = None,
        deferrable: Deferrable | None = None,
        include: Sequence[str] | None = None,
        opclasses: Sequence[Any] = (),
        nulls_distinct: bool | None = None,
        violation_error_code: str | None = None,
        violation_error_message: _StrOrPromise | None = None,
    ) -> None: ...
    @property
    @override
    def contains_expressions(self) -> bool: ...
    @override
    def check(self, model: type[Model], connection: BaseDatabaseWrapper) -> list[CheckMessage]: ...
    @override
    def validate(
        self, model: type[Model], instance: Model, exclude: set[str] | None = None, using: str = "default"
    ) -> None: ...

__all__ = ["BaseConstraint", "CheckConstraint", "Deferrable", "UniqueConstraint"]



import getpass as getpass
from typing import Any

from django.contrib.auth.base_user import AbstractBaseUser
from django.core.management.base import BaseCommand
from django.db.models import Field
from django.utils.functional import cached_property

class NotRunningInTTYException(Exception): ...

PASSWORD_FIELD: str

class Command(BaseCommand):
    UserModel: type[AbstractBaseUser]
    username_field: Field
    stdin: Any
    def get_input_data(self, field: Field, message: str, default: str | None = ...) -> str | None: ...
    @cached_property
    def username_is_unique(self) -> bool: ...



from collections import deque
from collections.abc import Iterable
from functools import cached_property
from typing import Any, cast

from django.db.models import Manager, Model
from django.db.models.fields import DateField, DateTimeField, Field
from django.db.models.fields.reverse_related import ForeignObjectRel, ManyToManyRel, OneToOneRel
from mypy.checker import TypeChecker
from mypy.nodes import (
    ARG_STAR2,
    MDEF,
    Argument,
    AssignmentStmt,
    CallExpr,
    Context,
    FakeInfo,
    NameExpr,
    RefExpr,
    Statement,
    SymbolTableNode,
    TypeInfo,
    Var,
)
from mypy.plugin import AnalyzeTypeContext, AttributeContext, ClassDefContext
from mypy.plugins import common
from mypy.semanal import SemanticAnalyzer
from mypy.typeanal import TypeAnalyser
from mypy.types import AnyType, Instance, ProperType, TypedDictType, TypeOfAny, TypeType, TypeVarType, get_proper_type
from mypy.types import Type as MypyType
from mypy.typevars import fill_typevars, fill_typevars_with_any
from typing_extensions import override

from mypy_django_plugin.config import DjangoPluginConfig
from mypy_django_plugin.django.context import DjangoContext
from mypy_django_plugin.errorcodes import MANAGER_MISSING
from mypy_django_plugin.exceptions import UnregisteredModelError
from mypy_django_plugin.lib import fullnames, helpers
from mypy_django_plugin.lib.fullnames import ANNOTATIONS_FULLNAME
from mypy_django_plugin.transformers.fields import FieldDescriptorTypes, get_field_descriptor_types
from mypy_django_plugin.transformers.managers import (
    MANAGER_METHODS_RETURNING_QUERYSET,
    create_manager_info_from_from_queryset_call,
)
from mypy_django_plugin.transformers.manytomany import M2MArguments, M2MThrough, M2MTo


class ModelClassInitializer:
    api: SemanticAnalyzer

    def __init__(self, ctx: ClassDefContext, django_context: DjangoContext) -> None:
        self.api = cast("SemanticAnalyzer", ctx.api)
        self.model_classdef = ctx.cls
        self.django_context = django_context
        self.ctx = ctx

    @property
    def is_model_abstract(self) -> bool:
        return helpers.is_abstract_model(self.model_classdef.info)

    def lookup_typeinfo(self, fullname: str) -> TypeInfo | None:
        return helpers.lookup_fully_qualified_typeinfo(self.api, fullname)

    def lookup_typeinfo_or_incomplete_defn_error(self, fullname: str) -> TypeInfo:
        info = self.lookup_typeinfo(fullname)
        if info is None:
            raise helpers.IncompleteDefnException(f"No {fullname!r} found")
        return info

    def lookup_class_typeinfo_or_incomplete_defn_error(self, klass: type) -> TypeInfo:
        fullname = helpers.get_class_fullname(klass)
        return self.lookup_typeinfo_or_incomplete_defn_error(fullname)

    def create_new_var(self, name: str, typ: MypyType) -> Var:
        # type=: type of the variable itself
        var = Var(name=name, type=typ)
        # var.info: type of the object variable is bound to
        var.info = self.model_classdef.info
        var._fullname = self.model_classdef.info.fullname + "." + name
        var.is_initialized_in_class = True
        var.is_inferred = True
        return var

    def add_new_node_to_model_class(
        self, name: str, typ: MypyType, *, no_serialize: bool = False, is_classvar: bool = False
    ) -> None:
        # TODO: Rename to signal that it is a `Var` that is added..
        helpers.add_new_sym_for_info(
            self.model_classdef.info, name=name, sym_type=typ, no_serialize=no_serialize, is_classvar=is_classvar
        )

    def add_new_class_for_current_module(self, name: str, bases: list[Instance]) -> TypeInfo:
        current_module = self.api.modules[self.model_classdef.info.module_name]
        return helpers.add_new_class_for_module(current_module, name=name, bases=bases)

    def run(self) -> None:
        model_cls = self.django_context.get_model_class_by_fullname(self.model_classdef.fullname)
        if model_cls is None:
            return
        self.run_with_model_cls(model_cls)

    def get_generated_manager_mappings(self, base_manager_fullname: str) -> dict[str, str]:
        base_manager_info = self.lookup_typeinfo(base_manager_fullname)
        if base_manager_info is None or "from_queryset_managers" not in base_manager_info.metadata:
            return {}
        return base_manager_info.metadata["from_queryset_managers"]

    def get_generated_manager_info(self, manager_fullname: str, base_manager_fullname: str) -> TypeInfo | None:
        generated_managers = self.get_generated_manager_mappings(base_manager_fullname)
        real_manager_fullname = generated_managers.get(manager_fullname)
        if real_manager_fullname:
            return self.lookup_typeinfo(real_manager_fullname)
        # Not a generated manager
        return None

    def get_or_create_manager_with_any_fallback(self) -> TypeInfo | None:
        """
        Create a Manager subclass with fallback to Any for unknown attributes
        and methods. This is used for unresolved managers, where we don't know
        the actual type of the manager.

        The created class is reused if multiple unknown managers are encountered.
        """

        name = "UnknownManager"

        # Check if we've already created a fallback manager class for this
        # module, and if so reuse that.
        manager_info = self.lookup_typeinfo(f"{self.model_classdef.info.module_name}.{name}")
        if manager_info and manager_info.metadata.get("django", {}).get("any_fallback_manager"):
            return manager_info

        fallback_queryset = self.get_or_create_queryset_with_any_fallback()
        if fallback_queryset is None:
            return None
        base_manager_fullname = fullnames.MANAGER_CLASS_FULLNAME
        base_manager_info = self.lookup_typeinfo(base_manager_fullname)
        if base_manager_info is None:
            return None

        base_manager = fill_typevars(base_manager_info)
        assert isinstance(base_manager, Instance)
        manager_info = self.add_new_class_for_current_module(name, [base_manager])
        manager_info.fallback_to_any = True

        manager_info.type_vars = base_manager_info.type_vars
        manager_info.defn.type_vars = base_manager_info.defn.type_vars
        manager_info.metaclass_type = manager_info.calculate_metaclass_type()

        # For methods on BaseManager that return a queryset we need to update
        # the return type to be the actual queryset subclass used. This is done
        # by adding the methods as attributes with type Any to the manager
        # class. The actual type of these methods are resolved in
        # resolve_manager_method.
        for method_name in MANAGER_METHODS_RETURNING_QUERYSET:
            helpers.add_new_sym_for_info(
                manager_info, name=method_name, sym_type=AnyType(TypeOfAny.implementation_artifact)
            )

        manager_info.metadata["django"] = {
            "any_fallback_manager": True,
            "from_queryset_manager": fallback_queryset.fullname,
        }

        return manager_info

    def get_or_create_queryset_with_any_fallback(self) -> TypeInfo | None:
        """
        Create a QuerySet subclass with fallback to Any for unknown attributes
        and methods. This is used for the manager returned by the method above.
        """

        name = "UnknownQuerySet"

        # Check if we've already created a fallback queryset class for this
        # module, and if so reuse that.
        queryset_info = self.lookup_typeinfo(f"{self.model_classdef.info.module_name}.{name}")
        if queryset_info and queryset_info.metadata.get("django", {}).get("any_fallback_queryset"):
            return queryset_info

        base_queryset_info = self.lookup_typeinfo(fullnames.QUERYSET_CLASS_FULLNAME)
        if base_queryset_info is None:
            return None

        base_queryset = fill_typevars(base_queryset_info)
        assert isinstance(base_queryset, Instance)
        queryset_info = self.add_new_class_for_current_module(name, [base_queryset])
        queryset_info.metadata["django"] = {
            "any_fallback_queryset": True,
        }
        queryset_info.fallback_to_any = True

        queryset_info.type_vars = base_queryset_info.type_vars.copy()
        queryset_info.defn.type_vars = base_queryset_info.defn.type_vars.copy()
        queryset_info.metaclass_type = queryset_info.calculate_metaclass_type()

        return queryset_info

    def run_with_model_cls(self, model_cls: type[Model]) -> None:
        raise NotImplementedError(f"Implement this in subclass {self.__class__.__name__}")


class AddAnnotateUtilities(ModelClassInitializer):
    """
    Creates a model subclass that will be used when the model's manager/queryset calls
    'annotate' to hold on to attributes that Django adds to a model instance.

    Example:

        class MyModel(models.Model):
            ...

        class MyModel@AnnotatedWith(MyModel, django_stubs_ext.Annotations[_Annotations]):
            ...
    """

    @override
    def run(self) -> None:
        annotations = self.lookup_typeinfo_or_incomplete_defn_error("django_stubs_ext.Annotations")
        exception_bases = {
            model_exc_name: self.lookup_typeinfo_or_incomplete_defn_error(base_exc_name)
            for model_exc_name, base_exc_name in [
                ("DoesNotExist", fullnames.OBJECT_DOES_NOT_EXIST),
                ("NotUpdated", fullnames.OBJECT_NOT_UPDATED),
                ("MultipleObjectsReturned", fullnames.MULTIPLE_OBJECTS_RETURNED),
            ]
        }
        annotated_model_name = self.model_classdef.info.name + "@AnnotatedWith"
        annotated_model = self.lookup_typeinfo(self.model_classdef.info.module_name + "." + annotated_model_name)
        if annotated_model is None:
            model_type = fill_typevars_with_any(self.model_classdef.info)
            assert isinstance(model_type, Instance)
            annotations_type = fill_typevars(annotations)
            assert isinstance(annotations_type, Instance)
            annotated_model = self.add_new_class_for_current_module(
                annotated_model_name, bases=[model_type, annotations_type]
            )
            annotated_model.defn.type_vars = annotations.defn.type_vars
            annotated_model.add_type_vars()
            helpers.mark_as_annotated_model(annotated_model)
            if self.is_model_abstract:
                # Below are abstract attributes, and in a stub file mypy requires
                # explicit ABCMeta if not all abstract attributes are implemented i.e.
                # class is kept abstract. So we add the attributes to get mypy off our
                # back
                for model_exc_name, base_exc_type in exception_bases.items():
                    helpers.add_new_sym_for_info(
                        annotated_model,
                        model_exc_name,
                        TypeType(Instance(base_exc_type, [])),
                    )


class InjectAnyAsBaseForNestedMeta(ModelClassInitializer):
    """
    Replaces
        class MyModel(models.Model):
            class Meta:
                pass
    with
        class MyModel(models.Model):
            class Meta(TypedModelMeta):
                pass

    to provide proper typing of attributes in Meta inner classes.

    If TypedModelMeta is not available, fallback to Any as a base
    to get around incompatible Meta inner classes for different models.
    """

    @override
    def run(self) -> None:
        meta_node = helpers.get_nested_meta_node_for_current_class(self.model_classdef.info)
        if meta_node is None:
            return None
        meta_node.fallback_to_any = True

        typed_model_meta_info = self.lookup_typeinfo(fullnames.TYPED_MODEL_META_FULLNAME)
        if typed_model_meta_info and not meta_node.has_base(fullnames.TYPED_MODEL_META_FULLNAME):
            # Insert TypedModelMeta just before `object` to leverage mypy's class-body semantic analysis.
            meta_node.mro.insert(-1, typed_model_meta_info)
        return None


class AddDefaultPrimaryKey(ModelClassInitializer):
    @override
    def run_with_model_cls(self, model_cls: type[Model]) -> None:
        auto_field = model_cls._meta.auto_field
        if auto_field:
            self.create_autofield(
                auto_field=auto_field,
                dest_name=auto_field.attname,
                existing_field=not self.model_classdef.info.has_readable_member(auto_field.attname),
            )

    def create_autofield(
        self,
        auto_field: "Field[Any, Any]",
        dest_name: str,
        existing_field: bool,
    ) -> None:
        if existing_field:
            auto_field_fullname = helpers.get_class_fullname(auto_field.__class__)
            auto_field_info = self.lookup_typeinfo_or_incomplete_defn_error(auto_field_fullname)

            set_type, get_type = get_field_descriptor_types(
                auto_field_info,
                is_set_nullable=True,
                is_get_nullable=False,
            )

            self.add_new_node_to_model_class(dest_name, Instance(auto_field_info, [set_type, get_type]))


class AddPrimaryKeyAlias(AddDefaultPrimaryKey):
    @override
    def run_with_model_cls(self, model_cls: type[Model]) -> None:
        # We also need to override existing `pk` definition from `stubs`:
        auto_field = model_cls._meta.pk
        if auto_field is not None:
            self.create_autofield(
                auto_field=auto_field,
                dest_name="pk",
                existing_field=self.model_classdef.info.has_readable_member(auto_field.name),
            )


class AddRelatedModelsId(ModelClassInitializer):
    @override
    def run_with_model_cls(self, model_cls: type[Model]) -> None:
        for field in self.django_context.get_model_foreign_keys(model_cls):
            try:
                related_model_cls = self.django_context.get_field_related_model_cls(field)
            except UnregisteredModelError:
                error_context: Context = self.ctx.cls
                field_sym = self.ctx.cls.info.get(field.name)
                if field_sym is not None and field_sym.node is not None:
                    error_context = field_sym.node
                self.api.fail(
                    f"Cannot find model {field.related_model!r} referenced in field {field.name!r}",
                    ctx=error_context,
                )
                self.add_new_node_to_model_class(field.attname, AnyType(TypeOfAny.explicit))
                continue

            if related_model_cls._meta.abstract:
                continue

            rel_target_field = self.django_context.get_related_target_field(related_model_cls, field)
            if not rel_target_field:
                continue

            try:
                field_info = self.lookup_class_typeinfo_or_incomplete_defn_error(rel_target_field.__class__)
            except helpers.IncompleteDefnException as exc:
                if not self.api.final_iteration:
                    raise exc
                continue

            is_nullable = self.django_context.get_field_nullability(field, None)
            set_type, get_type = get_field_descriptor_types(
                field_info, is_set_nullable=is_nullable, is_get_nullable=is_nullable
            )
            self.add_new_node_to_model_class(field.attname, Instance(field_info, [set_type, get_type]))


class AddManagers(ModelClassInitializer):
    def lookup_manager(self, fullname: str, manager: "Manager[Any]") -> TypeInfo | None:
        manager_info = self.lookup_typeinfo(fullname)
        if manager_info is None:
            manager_info = self.get_dynamic_manager(fullname, manager)
        return manager_info

    def is_manager_dynamically_generated(self, manager_info: TypeInfo | None) -> bool:
        if manager_info is None:
            return False
        return manager_info.metadata.get("django", {}).get("from_queryset_manager") is not None

    def reparametrize_dynamically_created_manager(self, manager_name: str, manager_info: TypeInfo | None) -> None:
        if not self.is_manager_dynamically_generated(manager_info):
            return

        assert manager_info is not None
        # Reparameterize dynamically created manager with model type
        manager_type = helpers.fill_manager(manager_info, Instance(self.model_classdef.info, []))
        self.add_new_node_to_model_class(manager_name, manager_type, is_classvar=True)

    @override
    def run_with_model_cls(self, model_cls: type[Model]) -> None:
        manager_info: TypeInfo | None

        incomplete_manager_defs = set()
        for manager_name, manager in model_cls._meta.managers_map.items():
            manager_node = self.model_classdef.info.get(manager_name)
            manager_fullname = helpers.get_class_fullname(manager.__class__)
            manager_info = self.lookup_manager(manager_fullname, manager)

            if manager_node and manager_node.type is not None:
                # Manager is already typed -> do nothing unless it's a dynamically generated manager
                self.reparametrize_dynamically_created_manager(manager_name, manager_info)
                continue

            if manager_info is None:
                # We couldn't find a manager type, see if we should create one
                manager_info = self.create_manager_from_from_queryset(manager_name)

            if manager_info is None:
                incomplete_manager_defs.add(manager_name)
                continue

            assert self.model_classdef.info.self_type is not None
            manager_type = helpers.fill_manager(manager_info, self.model_classdef.info.self_type)
            self.add_new_node_to_model_class(manager_name, manager_type, is_classvar=True)

        if incomplete_manager_defs:
            if not self.api.final_iteration:
                #  Unless we're on the final round, see if another round could
                #  figure out all manager types
                raise helpers.IncompleteDefnException()

            for manager_name in incomplete_manager_defs:
                # We act graceful and set the type as the bare minimum we know of
                # (Django's default) before finishing. And emit an error, to allow for
                # ignoring a more specialised manager not being resolved while still
                # setting _some_ type
                fallback_manager_info = self.get_or_create_manager_with_any_fallback()
                if fallback_manager_info is not None:
                    assert self.model_classdef.info.self_type is not None
                    manager_type = helpers.fill_manager(fallback_manager_info, self.model_classdef.info.self_type)
                    self.add_new_node_to_model_class(manager_name, manager_type, is_classvar=True)

                # Find expression for e.g. `objects = SomeManager()`
                manager_expr = self.get_manager_expression(manager_name)
                manager_fullname = f"{self.model_classdef.fullname}.{manager_name}"
                self.api.fail(
                    f'Could not resolve manager type for "{manager_fullname}"',
                    manager_expr or self.ctx.cls,
                    code=MANAGER_MISSING,
                )

    def get_manager_expression(self, name: str) -> AssignmentStmt | None:
        # TODO: What happens if the manager is defined multiple times?
        for expr in self.ctx.cls.defs.body:
            if (
                isinstance(expr, AssignmentStmt)
                and isinstance(expr.lvalues[0], NameExpr)
                and expr.lvalues[0].name == name
            ):
                return expr

        return None

    def get_dynamic_manager(self, fullname: str, manager: "Manager[Any]") -> TypeInfo | None:
        """
        Try to get a dynamically defined manager
        """

        # Check if manager is a generated (dynamic class) manager
        base_manager_fullname = helpers.get_class_fullname(manager.__class__.__bases__[0])
        generated_managers = self.get_generated_manager_mappings(base_manager_fullname)

        generated_manager_name: str | None = generated_managers.get(fullname, None)
        if generated_manager_name is None:
            return None

        return self.lookup_typeinfo(generated_manager_name)

    def create_manager_from_from_queryset(self, name: str) -> TypeInfo | None:
        """
        Try to create a manager from a .from_queryset call:

            class MyModel(models.Model):
                objects = MyManager.from_queryset(MyQuerySet)()
        """

        assign_statement = self.get_manager_expression(name)
        if assign_statement is None:
            return None

        expr = assign_statement.rvalue
        if not isinstance(expr, CallExpr) or not isinstance(expr.callee, CallExpr):
            return None

        return create_manager_info_from_from_queryset_call(self.api, expr.callee)


class AddDefaultManagerAttribute(ModelClassInitializer):
    @override
    def run_with_model_cls(self, model_cls: type[Model]) -> None:
        if "_default_manager" in self.model_classdef.info.names:
            return None

        default_manager_cls = model_cls._meta.default_manager.__class__
        default_manager_fullname = helpers.get_class_fullname(default_manager_cls)

        try:
            default_manager_info = self.lookup_typeinfo_or_incomplete_defn_error(default_manager_fullname)
        except helpers.IncompleteDefnException as exc:
            # Check if default manager could be a generated manager
            base_manager_fullname = helpers.get_class_fullname(default_manager_cls.__bases__[0])
            generated_manager_info = self.get_generated_manager_info(default_manager_fullname, base_manager_fullname)
            if generated_manager_info is None:
                # Manager doesn't appear to be generated. Unless we're on the final round,
                # see if another round could help figuring out the default manager type
                if not self.api.final_iteration:
                    raise exc
                return None
            default_manager_info = generated_manager_info

        default_manager = helpers.fill_manager(default_manager_info, Instance(self.model_classdef.info, []))
        self.add_new_node_to_model_class("_default_manager", default_manager, is_classvar=True)


class AddReverseLookups(ModelClassInitializer):
    @cached_property
    def reverse_one_to_one_descriptor(self) -> TypeInfo:
        return self.lookup_typeinfo_or_incomplete_defn_error(fullnames.REVERSE_ONE_TO_ONE_DESCRIPTOR)

    @cached_property
    def reverse_many_to_one_descriptor(self) -> TypeInfo:
        return self.lookup_typeinfo_or_incomplete_defn_error(fullnames.REVERSE_MANY_TO_ONE_DESCRIPTOR)

    @cached_property
    def many_to_many_descriptor(self) -> TypeInfo:
        return self.lookup_typeinfo_or_incomplete_defn_error(fullnames.MANY_TO_MANY_DESCRIPTOR)

    def process_relation(self, relation: ForeignObjectRel) -> None:
        attname = relation.get_accessor_name()
        if attname is None:
            # No reverse accessor.
            return

        to_model_cls = self.django_context.get_field_related_model_cls(relation)
        to_model_info = self.lookup_class_typeinfo_or_incomplete_defn_error(to_model_cls)

        reverse_lookup_declared = attname in self.model_classdef.info.names
        if isinstance(relation, OneToOneRel):
            if not reverse_lookup_declared:
                self.add_new_node_to_model_class(
                    attname,
                    Instance(
                        self.reverse_one_to_one_descriptor,
                        [Instance(self.model_classdef.info, []), Instance(to_model_info, [])],
                    ),
                )
            return
        if isinstance(relation, ManyToManyRel):
            if not reverse_lookup_declared:
                # TODO: 'relation' should be based on `TypeInfo` instead of Django runtime.
                assert relation.through is not None
                through_fullname = helpers.get_class_fullname(relation.through)
                through_model_info = self.lookup_typeinfo_or_incomplete_defn_error(through_fullname)
                self.add_new_node_to_model_class(
                    attname,
                    Instance(
                        self.many_to_many_descriptor, [Instance(to_model_info, []), Instance(through_model_info, [])]
                    ),
                    is_classvar=True,
                )
            return
        if not reverse_lookup_declared:
            # ManyToOneRel
            self.add_new_node_to_model_class(
                attname, Instance(self.reverse_many_to_one_descriptor, [Instance(to_model_info, [])]), is_classvar=True
            )

        related_manager_info = self.lookup_typeinfo_or_incomplete_defn_error(fullnames.RELATED_MANAGER_CLASS)
        # TODO: Support other reverse managers than `_default_manager`
        default_manager = to_model_info.names.get("_default_manager")
        if default_manager is None:
            if not self.api.final_iteration:
                raise helpers.IncompleteDefnException()
            # When we get no default manager we can't customize the reverse manager any
            # further and will just fall back to the manager declared on the descriptor.
            # If a django model has a Manager class that cannot be resolved statically
            # (if it is generated in a way where we cannot import it, like
            # `objects = my_manager_factory()`)
            #
            # See https://github.com/typeddjango/django-stubs/pull/993
            # for more information on when this error can occur.
            self.ctx.api.fail(
                (
                    f"Couldn't resolve related manager {attname!r} for relation "
                    f"'{to_model_info.fullname}.{relation.field.name}'."
                ),
                self.ctx.cls,
                code=MANAGER_MISSING,
            )
            return

        default_manager_type = get_proper_type(default_manager.type)
        if (
            # '_default_manager' attribute is a node type we can't process
            not isinstance(default_manager_type, Instance)
            # Already has a related manager subclassed from the default manager
            or helpers.get_reverse_manager_info(self.api, model_info=to_model_info, derived_from="_default_manager")
            is not None
            # When the default manager isn't custom there's no need to create a new type
            # as `RelatedManager` has `models.Manager` as base
            or default_manager_type.type.fullname == fullnames.MANAGER_CLASS_FULLNAME
        ):
            return

        # Create a reverse manager subclassed from the default manager of the related
        # model and 'RelatedManager'
        related_manager = Instance(related_manager_info, [Instance(to_model_info, [])])
        # The reverse manager is based on the related model's manager, so it makes most sense to add the new
        # related manager in that module
        new_related_manager_info = helpers.add_new_class_for_module(
            module=self.api.modules[to_model_info.module_name],
            name=f"{to_model_info.name}_RelatedManager",
            bases=[related_manager, default_manager_type],
        )
        # Stash the new reverse manager type fullname on the related model, so we don't duplicate
        # or have to create it again for other reverse relations
        helpers.set_reverse_manager_info(
            to_model_info,
            derived_from="_default_manager",
            fullname=new_related_manager_info.fullname,
        )

    @override
    def run_with_model_cls(self, model_cls: type[Model]) -> None:
        # add related managers etc.
        processing_incomplete = False
        for relation in self.django_context.get_model_relations(model_cls):
            try:
                self.process_relation(relation)
            except helpers.IncompleteDefnException:
                processing_incomplete = True

        if processing_incomplete and not self.api.final_iteration:
            raise helpers.IncompleteDefnException


class AddExtraFieldMethods(ModelClassInitializer):
    @override
    def run_with_model_cls(self, model_cls: type[Model]) -> None:
        # get_FOO_display for choices
        for field in self.django_context.get_model_fields(model_cls):
            if field.choices:
                info = self.lookup_typeinfo_or_incomplete_defn_error("builtins.str")
                return_type = Instance(info, [])
                common.add_method(self.ctx, name=f"get_{field.attname}_display", args=[], return_type=return_type)

        # get_next_by, get_previous_by for Date, DateTime
        for field in self.django_context.get_model_fields(model_cls):
            if isinstance(field, DateField | DateTimeField) and not field.null:
                return_type = Instance(self.model_classdef.info, [])
                common.add_method(
                    self.ctx,
                    name=f"get_next_by_{field.attname}",
                    args=[
                        Argument(
                            Var("kwargs", AnyType(TypeOfAny.implementation_artifact)),
                            AnyType(TypeOfAny.implementation_artifact),
                            initializer=None,
                            kind=ARG_STAR2,
                        )
                    ],
                    return_type=return_type,
                )
                common.add_method(
                    self.ctx,
                    name=f"get_previous_by_{field.attname}",
                    args=[
                        Argument(
                            Var("kwargs", AnyType(TypeOfAny.implementation_artifact)),
                            AnyType(TypeOfAny.implementation_artifact),
                            initializer=None,
                            kind=ARG_STAR2,
                        )
                    ],
                    return_type=return_type,
                )


class ProcessManyToManyFields(ModelClassInitializer):
    """
    Processes 'ManyToManyField()' fields and;

    - Generates any implicit through tables that Django also generates. It won't do
      anything if the model is abstract or for fields where an explicit 'through'
      argument has been passed.
    - Creates related managers for both ends of the many to many relationship

    TODO: Move the 'related_name' contribution from 'AddReverseLookups' to here. As it
          makes sense to add it when processing ManyToManyField
    """

    def statements(self) -> Iterable[Statement]:
        """
        Returns class body statements from the current model and any of its bases that
        is an abstract model. Statements from any concrete parent class or parents of
        that concrete class will be skipped.
        """
        processed_models = set()
        # Produce all statements from current class
        model_bases = deque([self.model_classdef])
        # Do a breadth first search over the current model and its bases, to find all
        # abstract parent models that have not been "interrupted" by any concrete model.
        while model_bases:
            model = model_bases.popleft()
            yield from model.defs.body
            if isinstance(model.info, FakeInfo):
                # While loading from cache ClassDef infos are faked and 'FakeInfo' doesn't have
                # all attributes of a 'TypeInfo' set. See #2184
                continue
            for base in model.info.bases:
                # Only produce any additional statements from abstract model bases, as they
                # simulate regular python inheritance. Avoid concrete models, and any of their
                # parents, as they're handled differently by Django.
                if helpers.is_abstract_model(base.type) and base.type.fullname not in processed_models:
                    model_bases.append(base.type.defn)
                    processed_models.add(base.type.fullname)

    @override
    def run(self) -> None:
        if self.is_model_abstract:
            # TODO: Create abstract through models?
            return

        for statement in self.statements():
            # Check if this part of the class body is an assignment from a 'ManyToManyField' call
            # <field> = ManyToManyField(...)
            if (
                isinstance(statement, AssignmentStmt)
                and len(statement.lvalues) == 1
                and isinstance(statement.lvalues[0], NameExpr)
                and isinstance(statement.rvalue, CallExpr)
                and len(statement.rvalue.args) > 0  # Need at least the 'to' argument
                and isinstance(statement.rvalue.callee, RefExpr)
                and isinstance(statement.rvalue.callee.node, TypeInfo)
                and statement.rvalue.callee.node.has_base(fullnames.MANYTOMANY_FIELD_FULLNAME)
            ):
                m2m_field_name = statement.lvalues[0].name
                m2m_field_symbol = self.model_classdef.info.get(m2m_field_name)
                # The symbol referred to by the assignment expression is expected to be a variable
                if m2m_field_symbol is None or not isinstance(m2m_field_symbol.node, Var):
                    continue
                # Resolve argument information of the 'ManyToManyField(...)' call
                args = self.resolve_many_to_many_arguments(statement.rvalue, context=statement)
                # Ignore calls without required 'to' argument, mypy will complain
                if args is None:
                    continue
                # Get the names of the implicit through model that will be generated
                through_model_name = f"{self.model_classdef.name}_{m2m_field_name}"
                through_model = self.create_through_table_class(
                    field_name=m2m_field_name,
                    model_name=through_model_name,
                    model_fullname=f"{self.model_classdef.info.module_name}.{through_model_name}",
                    m2m_args=args,
                )
                container = self.model_classdef.info.get_containing_type_info(m2m_field_name)
                if (
                    through_model is not None
                    and container is not None
                    and container.fullname != self.model_classdef.info.fullname
                    and helpers.is_abstract_model(container)
                ):
                    # ManyToManyField is inherited from an abstract parent class, so in
                    # order to get the to and the through model argument right we
                    # override the ManyToManyField attribute on the current class
                    helpers.add_new_sym_for_info(
                        self.model_classdef.info,
                        name=m2m_field_name,
                        sym_type=Instance(self.m2m_field, [args.to.model, Instance(through_model, [])]),
                    )
                # Create a 'ManyRelatedManager' class for the processed model
                self.create_many_related_manager(Instance(self.model_classdef.info, []))
                if isinstance(args.to.model, Instance):
                    # Create a 'ManyRelatedManager' class for the related model
                    self.create_many_related_manager(args.to.model)

    @cached_property
    def default_pk_instance(self) -> Instance:
        default_pk_field = self.lookup_typeinfo(self.django_context.settings.DEFAULT_AUTO_FIELD)
        if default_pk_field is None:
            raise helpers.IncompleteDefnException()
        return Instance(
            default_pk_field,
            list(get_field_descriptor_types(default_pk_field, is_set_nullable=True, is_get_nullable=False)),
        )

    @cached_property
    def model_pk_instance(self) -> Instance:
        return self.get_pk_instance(self.model_classdef.info)

    @cached_property
    def model_base(self) -> TypeInfo:
        info = self.lookup_typeinfo(fullnames.MODEL_CLASS_FULLNAME)
        if info is None:
            raise helpers.IncompleteDefnException()
        return info

    @cached_property
    def fk_field(self) -> TypeInfo:
        info = self.lookup_typeinfo(fullnames.FOREIGN_KEY_FULLNAME)
        if info is None:
            raise helpers.IncompleteDefnException()
        return info

    @cached_property
    def m2m_field(self) -> TypeInfo:
        info = self.lookup_typeinfo(fullnames.MANYTOMANY_FIELD_FULLNAME)
        if info is None:
            raise helpers.IncompleteDefnException()
        return info

    @cached_property
    def manager_info(self) -> TypeInfo:
        info = self.lookup_typeinfo(fullnames.MANAGER_CLASS_FULLNAME)
        if info is None:
            raise helpers.IncompleteDefnException()
        return info

    @cached_property
    def fk_field_types(self) -> FieldDescriptorTypes:
        return get_field_descriptor_types(self.fk_field, is_set_nullable=False, is_get_nullable=False)

    @cached_property
    def many_related_manager(self) -> TypeInfo:
        return self.lookup_typeinfo_or_incomplete_defn_error(fullnames.MANY_RELATED_MANAGER)

    def get_pk_instance(self, model: TypeInfo, /) -> Instance:
        """
        Get a primary key instance of provided model's type info. If primary key can't be resolved,
        return a default declaration.
        """
        contains_from_pk_info = model.get_containing_type_info("pk")
        if contains_from_pk_info is not None:
            pk = contains_from_pk_info.names["pk"].node
            if isinstance(pk, Var):
                pk_type = get_proper_type(pk.type)
                if isinstance(pk_type, Instance):
                    return pk_type
        return self.default_pk_instance

    def create_through_table_class(
        self, field_name: str, model_name: str, model_fullname: str, m2m_args: M2MArguments
    ) -> TypeInfo | None:
        if not isinstance(m2m_args.to.model, Instance):
            return None
        if m2m_args.through is not None:
            # Call has explicit 'through=', no need to create any implicit through table
            return m2m_args.through.model.type if isinstance(m2m_args.through.model, Instance) else None

        # If through model is already declared there's nothing more we should do
        through_model = self.lookup_typeinfo(model_fullname)
        if through_model is not None:
            return through_model
        # Declare a new, empty, implicitly generated through model class named: '<Model>_<field_name>'
        through_model = self.add_new_class_for_current_module(model_name, bases=[Instance(self.model_base, [])])
        # We attempt to be a bit clever here and store the generated through model's fullname in
        # the metadata of the class containing the 'ManyToManyField' call expression, where its
        # identifier is the field name of the 'ManyToManyField'. This would allow the containing
        # model to always find the implicit through model, so that it doesn't get lost.
        model_metadata = helpers.get_django_metadata(self.model_classdef.info)
        model_metadata.setdefault("m2m_throughs", {})
        model_metadata["m2m_throughs"][field_name] = through_model.fullname
        # Add a 'pk' symbol to the model class
        helpers.add_new_sym_for_info(through_model, name="pk", sym_type=self.default_pk_instance.copy_modified())
        # Add an 'id' symbol to the model class
        helpers.add_new_sym_for_info(through_model, name="id", sym_type=self.default_pk_instance.copy_modified())
        # Add the foreign key to the model containing the 'ManyToManyField' call:
        # <containing_model> or from_<model>
        from_name = f"from_{self.model_classdef.name.lower()}" if m2m_args.to.self else self.model_classdef.name.lower()
        helpers.add_new_sym_for_info(
            through_model,
            name=from_name,
            sym_type=Instance(
                self.fk_field,
                [
                    helpers.convert_any_to_type(self.fk_field_types.set, Instance(self.model_classdef.info, [])),
                    helpers.convert_any_to_type(self.fk_field_types.get, Instance(self.model_classdef.info, [])),
                ],
            ),
        )
        # Add the foreign key's '_id' field: <containing_model>_id or from_<model>_id
        helpers.add_new_sym_for_info(
            through_model, name=f"{from_name}_id", sym_type=self.model_pk_instance.copy_modified()
        )
        # Add the foreign key to the model on the opposite side of the relation
        # i.e. the model given as 'to' argument to the 'ManyToManyField' call:
        # <other_model> or to_<model>
        to_name = (
            f"to_{m2m_args.to.model.type.name.lower()}" if m2m_args.to.self else m2m_args.to.model.type.name.lower()
        )
        helpers.add_new_sym_for_info(
            through_model,
            name=to_name,
            sym_type=Instance(
                self.fk_field,
                [
                    helpers.convert_any_to_type(self.fk_field_types.set, m2m_args.to.model),
                    helpers.convert_any_to_type(self.fk_field_types.get, m2m_args.to.model),
                ],
            ),
        )
        # Add the foreign key's '_id' field: <other_model>_id or to_<model>_id
        other_pk = self.get_pk_instance(m2m_args.to.model.type)
        helpers.add_new_sym_for_info(through_model, name=f"{to_name}_id", sym_type=other_pk.copy_modified())
        # Add a manager named 'objects'
        helpers.add_new_sym_for_info(
            through_model,
            name="objects",
            sym_type=Instance(self.manager_info, [Instance(through_model, [])]),
            is_classvar=True,
        )
        # Also add manager as '_default_manager' attribute
        helpers.add_new_sym_for_info(
            through_model,
            name="_default_manager",
            sym_type=Instance(self.manager_info, [Instance(through_model, [])]),
            is_classvar=True,
        )
        return through_model

    def resolve_many_to_many_arguments(self, call: CallExpr, /, context: Context) -> M2MArguments | None:
        """
        Inspect a 'ManyToManyField(...)' call to collect argument data on any 'to' and
        'through' arguments.
        """
        to_arg = helpers.get_class_init_argument_by_name(call, "to")
        if to_arg is None:
            # 'to' is a required argument of 'ManyToManyField()', we can't do anything if it's not provided
            return None

        # Resolve the type of the 'to' argument expression
        to_model = helpers.get_model_from_expression(
            to_arg, self_model=self.model_classdef.info, api=self.api, django_context=self.django_context
        )
        if to_model is None:
            return None
        to = M2MTo(
            arg=to_arg,
            model=to_model,
            self=to_model.type == self.model_classdef.info,
        )

        # Resolve the type of the 'through' argument expression
        through_arg = helpers.get_class_init_argument_by_name(call, "through")
        through = None
        if through_arg is not None:
            through_model = helpers.get_model_from_expression(
                through_arg, self_model=self.model_classdef.info, api=self.api, django_context=self.django_context
            )
            if through_model is not None:
                through = M2MThrough(arg=through_arg, model=through_model)

        return M2MArguments(to=to, through=through)

    def create_many_related_manager(self, model: Instance) -> None:
        """
        Creates a generic manager that subclasses both 'ManyRelatedManager' and the
        default manager of the given model. These are normally used on both models
        involved in a ManyToManyField.

        The manager classes are generic over a '_Through' model, meaning that they can
        be reused for multiple many to many relations.
        """
        if helpers.get_many_to_many_manager_info(self.api, to=model.type, derived_from="_default_manager") is not None:
            return

        default_manager_node = model.type.names.get("_default_manager")
        if default_manager_node is None:
            raise helpers.IncompleteDefnException()
        default_manager_type = get_proper_type(default_manager_node.type)
        if not isinstance(default_manager_type, Instance):
            return

        # Create a reusable generic subclass that is generic over a 'through' model,
        # explicitly declared it'd could have looked something like below
        #
        # class X(models.Model): ...
        # _Through = TypeVar("_Through", bound=models.Model)
        # class X_ManyRelatedManager(ManyRelatedManager[X, _Through], type(X._default_manager), Generic[_Through]): ...
        through_type_var = self.many_related_manager.defn.type_vars[1]
        assert isinstance(through_type_var, TypeVarType)
        generic_to_many_related_manager = Instance(self.many_related_manager, [model, through_type_var.copy_modified()])
        related_manager_info = helpers.add_new_class_for_module(
            module=self.api.modules[model.type.module_name],
            name=f"{model.type.name}_ManyRelatedManager",
            bases=[generic_to_many_related_manager, default_manager_type],
        )
        # Reuse the '_Through' `TypeVar` from `ManyRelatedManager` in our subclass
        related_manager_info.defn.type_vars = [through_type_var.copy_modified()]
        related_manager_info.add_type_vars()
        # Track the existence of our manager subclass, by tying it to the model it
        # operates on
        helpers.set_many_to_many_manager_info(
            to=model.type, derived_from="_default_manager", manager_info=related_manager_info
        )


class MetaclassAdjustments(ModelClassInitializer):
    @classmethod
    def adjust_model_class(cls, ctx: ClassDefContext, plugin_config: DjangoPluginConfig) -> None:
        """
        For the sake of type checkers other than mypy, some attributes that are
        dynamically added by Django's model metaclass has been annotated on
        `django.db.models.base.Model`. We remove those attributes and will handle them
        through the plugin.

        Configurable with `strict_model_abstract_attrs = false` to skip removing any objects from models.

        Basically, this code::

            from django.db import models

            def get_model_count(model_type: type[models.Model]) -> int:
                return model_type.objects.count()

        - Will raise an error
          `"type[models.Model]" has no attribute "objects"  [attr-defined]`
          when `strict_model_abstract_attrs = true` (default)
        - Return without any type errors
          when `strict_model_abstract_attrs = false`

        But, beware that the code above can fail in runtime, as mypy correctly shows.
        Example: `get_model_count(models.Model)`

        Turn this setting off at your own risk.
        """
        if not plugin_config.strict_model_abstract_attrs:
            return
        if ctx.cls.fullname != fullnames.MODEL_CLASS_FULLNAME:
            return

        for attr_name in ["DoesNotExist", "NotUpdated", "MultipleObjectsReturned", "objects"]:
            attr = ctx.cls.info.names.get(attr_name)
            if attr is not None and isinstance(attr.node, Var) and not attr.plugin_generated:
                del ctx.cls.info.names[attr_name]

    def get_exception_bases(self, name: str) -> list[Instance]:
        bases = []
        for model_base in self.model_classdef.info.direct_base_classes():
            exception_base_sym = model_base.names.get(name)
            if (
                # Base class is a Model
                helpers.is_model_type(model_base)
                # But base class is not 'models.Model'
                and model_base.fullname != fullnames.MODEL_CLASS_FULLNAME
                # Base class also has a generated exception base e.g. 'DoesNotExist'
                and exception_base_sym is not None
                and exception_base_sym.plugin_generated
                and isinstance(exception_base_sym.node, TypeInfo)
            ):
                bases.append(Instance(exception_base_sym.node, []))

        return bases

    def add_exception_classes(self) -> None:
        """
        Adds exception classes `DoesNotExist`, `NotUpdated`,  and `MultipleObjectsReturned` to a
        model type, aligned with how the model metaclass does it runtime.

        If the model is abstract, exceptions will be added as abstract attributes.
        """
        for model_exc_name, base_exc_name in [
            ("DoesNotExist", fullnames.OBJECT_DOES_NOT_EXIST),
            ("NotUpdated", fullnames.OBJECT_NOT_UPDATED),
            ("MultipleObjectsReturned", fullnames.MULTIPLE_OBJECTS_RETURNED),
        ]:
            if model_exc_name in self.model_classdef.info.names:
                continue

            base_exc_type = self.lookup_typeinfo_or_incomplete_defn_error(base_exc_name)
            base_exc_inst = Instance(base_exc_type, [])
            model_exc_type: Var | TypeInfo

            if self.is_model_abstract:
                model_exc_type = self.create_new_var(model_exc_name, TypeType(base_exc_inst))
                model_exc_type.is_abstract_var = True
            else:
                model_exc_type = helpers.create_type_info(
                    model_exc_name,
                    self.model_classdef.info.fullname,
                    self.get_exception_bases(model_exc_name) or [base_exc_inst],
                )

            self.model_classdef.info.names[model_exc_type.name] = SymbolTableNode(
                MDEF, model_exc_type, plugin_generated=True
            )

    @override
    def run(self) -> None:
        self.add_exception_classes()


def process_model_class(ctx: ClassDefContext, django_context: DjangoContext) -> None:
    initializers = [
        AddAnnotateUtilities,
        InjectAnyAsBaseForNestedMeta,
        AddDefaultPrimaryKey,
        AddPrimaryKeyAlias,
        AddRelatedModelsId,
        AddManagers,
        AddDefaultManagerAttribute,
        AddReverseLookups,
        AddExtraFieldMethods,
        ProcessManyToManyFields,
        MetaclassAdjustments,
    ]
    for initializer_cls in initializers:
        try:
            initializer_cls(ctx, django_context).run()
        except helpers.IncompleteDefnException:
            if not ctx.api.final_iteration:
                ctx.api.defer()


def set_auth_user_model_boolean_fields(ctx: AttributeContext, django_context: DjangoContext) -> MypyType:
    boolinfo = helpers.lookup_class_typeinfo(helpers.get_typechecker_api(ctx), bool)
    assert boolinfo is not None
    return Instance(boolinfo, [])


def handle_annotated_type(ctx: AnalyzeTypeContext, fullname: str) -> MypyType:
    """
    Replaces the 'WithAnnotations' type with a type that can represent an annotated
    model.
    """
    is_with_annotations = fullname == fullnames.WITH_ANNOTATIONS_FULLNAME
    args = ctx.type.args
    if not args:
        return AnyType(TypeOfAny.from_omitted_generics) if is_with_annotations else ctx.type
    type_arg = get_proper_type(ctx.api.analyze_type(args[0]))
    if not isinstance(type_arg, Instance) or not helpers.is_model_type(type_arg.type):
        return type_arg

    fields_dict = None
    if len(args) > 1:
        second_arg_type = get_proper_type(ctx.api.analyze_type(args[1]))
        if isinstance(second_arg_type, TypedDictType) and is_with_annotations:
            fields_dict = second_arg_type
        elif isinstance(second_arg_type, Instance) and second_arg_type.type.fullname == ANNOTATIONS_FULLNAME:
            annotations_type_arg = get_proper_type(second_arg_type.args[0])
            if isinstance(annotations_type_arg, TypedDictType):
                fields_dict = annotations_type_arg
            elif not isinstance(annotations_type_arg, AnyType):
                ctx.api.fail("Only TypedDicts are supported as type arguments to Annotations", ctx.context)
            elif annotations_type_arg.type_of_any == TypeOfAny.from_omitted_generics:
                ctx.api.fail("Missing required TypedDict parameter for generic type Annotations", ctx.context)

    if fields_dict is None:
        return type_arg

    assert isinstance(ctx.api, TypeAnalyser)
    assert isinstance(ctx.api.api, SemanticAnalyzer)
    return get_annotated_type(ctx.api.api, type_arg, fields_dict=fields_dict)


def get_annotated_type(
    api: SemanticAnalyzer | TypeChecker, model_type: Instance, fields_dict: TypedDictType
) -> ProperType:
    """
    Get a model type that can be used to represent an annotated model
    """
    extra_attrs = helpers.merge_extra_attrs(model_type.extra_attrs, new_attrs=fields_dict.items)

    annotated_model: TypeInfo | None
    if helpers.is_annotated_model(model_type.type):
        annotated_model = model_type.type
        if model_type.args:
            annotations = get_proper_type(model_type.args[0])
            if isinstance(annotations, TypedDictType):
                fields_dict = helpers.make_typeddict(
                    api,
                    fields={**annotations.items, **fields_dict.items},
                    required_keys={*annotations.required_keys, *fields_dict.required_keys},
                    readonly_keys={*annotations.readonly_keys, *fields_dict.readonly_keys},
                )
    else:
        annotated_model = helpers.lookup_fully_qualified_typeinfo(api, model_type.type.fullname + "@AnnotatedWith")

    if annotated_model is None:
        return model_type
    return Instance(annotated_model, [fields_dict], extra_attrs=extra_attrs)



from typing import Any

from django.core.checks.messages import CheckMessage
from django.db.backends.base.validation import BaseDatabaseValidation
from django.db.backends.mysql.base import DatabaseWrapper
from django.db.models.fields import Field
from typing_extensions import override

class DatabaseValidation(BaseDatabaseValidation):
    connection: DatabaseWrapper
    @override
    def check(self, **kwargs: Any) -> list[CheckMessage]: ...
    def check_field_type(self, field: Field, field_type: str) -> list[CheckMessage]: ...



from typing import Any, ClassVar

from django.contrib.postgres.fields.array import ArrayField
from django.contrib.postgres.utils import CheckPostgresInstalledMixin
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models import Field, TextField, Transform
from django.db.models.fields.mixins import CheckFieldDefaultMixin
from django.db.models.sql.compiler import SQLCompiler, _AsSqlType
from typing_extensions import override

class HStoreField(CheckPostgresInstalledMixin, CheckFieldDefaultMixin, Field):
    @override
    def get_transform(self, name: str) -> Any: ...
    @override
    def formfield(self, **kwargs: Any) -> Any: ...  # type: ignore[override]

class KeyTransform(Transform):
    output_field: ClassVar[TextField]

    def __init__(self, key_name: str, *args: Any, **kwargs: Any) -> None: ...
    @override
    def as_sql(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...  # type: ignore[override]

class KeyTransformFactory:
    def __init__(self, key_name: str) -> None: ...
    def __call__(self, *args: Any, **kwargs: Any) -> KeyTransform: ...

class KeysTransform(Transform):
    output_field: ClassVar[ArrayField]

class ValuesTransform(Transform):
    output_field: ClassVar[ArrayField]

__all__ = ["HStoreField"]



from enum import IntEnum
from types import ModuleType
from typing import Any

from django.db.backends.base.base import BaseDatabaseWrapper

is_psycopg3: bool

TSRANGE_OID: int
TSTZRANGE_OID: int

errors: ModuleType
sql: ModuleType
adapt: ModuleType
adapters: Any

class IsolationLevel(IntEnum):
    READ_UNCOMMITTED = 1
    READ_COMMITTED = 2
    REPEATABLE_READ = 3
    SERIALIZABLE = 4

# Range types from psycopg
Range: Any
DateRange: Any
DateTimeRange: Any
DateTimeTZRange: Any
NumericRange: Any
RANGE_TYPES: tuple[Any, ...]

# Psycopg types
Jsonb: Any
ClientCursor: Any
TextLoader: Any
TimestamptzLoader: Any
BaseTzLoader: Any
RangeDumper: Any
DjangoRangeDumper: Any

def Inet(address: str) -> Any: ...
def mogrify(sql: str, params: Any, connection: BaseDatabaseWrapper) -> str: ...
def register_tzloader(tz: Any, context: Any) -> None: ...
def get_adapters_template(use_tz: bool, timezone: Any) -> Any: ...



from typing import Any

from django.db.migrations.migration import Migration
from django.db.migrations.operations.base import Operation
from django.db.migrations.serializer import BaseSerializer

class OperationWriter:
    operation: Operation
    buff: list[Any]
    indentation: int
    def __init__(self, operation: Operation, indentation: int = 2) -> None: ...
    def serialize(self) -> tuple[str, set[str]]: ...
    def indent(self) -> None: ...
    def unindent(self) -> None: ...
    def feed(self, line: str) -> None: ...
    def render(self) -> str: ...

class MigrationWriter:
    migration: Migration
    needs_manual_porting: bool
    def __init__(self, migration: Migration, include_header: bool = True) -> None: ...
    def as_string(self) -> str: ...
    @property
    def basedir(self) -> str: ...
    @property
    def filename(self) -> str: ...
    @property
    def path(self) -> str: ...
    @classmethod
    def serialize(cls, value: Any) -> tuple[str, set[str]]: ...
    @classmethod
    def register_serializer(cls, type_: type, serializer: type[BaseSerializer]) -> None: ...
    @classmethod
    def unregister_serializer(cls, type_: type) -> None: ...

MIGRATION_HEADER_TEMPLATE: str
MIGRATION_TEMPLATE: str



from typing import Any

UserModel: Any

def check_password(environ: dict[Any, Any], username: str, password: str) -> bool | None: ...
def groups_for_user(environ: dict[Any, Any], username: str) -> list[bytes]: ...



from collections.abc import Callable, Iterable
from types import TracebackType
from typing import Any, TypeVar

from django.apps import AppConfig
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models import Model
from django.utils.connection import BaseConnectionHandler
from django.utils.functional import cached_property
from typing_extensions import override

DEFAULT_DB_ALIAS: str
DJANGO_VERSION_PICKLE_KEY: str

class Error(Exception): ...
class InterfaceError(Error): ...
class DatabaseError(Error): ...
class DataError(DatabaseError): ...
class OperationalError(DatabaseError): ...
class IntegrityError(DatabaseError): ...
class InternalError(DatabaseError): ...
class ProgrammingError(DatabaseError): ...
class NotSupportedError(DatabaseError): ...

_F = TypeVar("_F", bound=Callable[..., Any])

class DatabaseErrorWrapper:
    def __init__(self, wrapper: Any) -> None: ...
    def __del__(self) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None: ...
    def __call__(self, func: _F) -> _F: ...

def load_backend(backend_name: str) -> Any: ...

class ConnectionHandler(BaseConnectionHandler[BaseDatabaseWrapper]):
    @override
    def configure_settings(self, databases: dict[str, Any] | None) -> dict[str, Any]: ...
    @property
    def databases(self) -> dict[str, dict[str, Any]]: ...
    @override
    def create_connection(self, alias: str) -> BaseDatabaseWrapper: ...

class ConnectionRouter:
    def __init__(self, routers: Iterable[Any] | None = None) -> None: ...
    @cached_property
    def routers(self) -> list[Any]: ...
    def db_for_read(self, model: type[Model], **hints: Any) -> str: ...
    def db_for_write(self, model: type[Model], **hints: Any) -> str: ...
    def allow_relation(self, obj1: Model, obj2: Model, **hints: Any) -> bool: ...
    def allow_migrate(self, db: str, app_label: str, **hints: Any) -> bool: ...
    def allow_migrate_model(self, db: str, model: type[Model]) -> bool: ...
    def get_migratable_models(
        self, app_config: AppConfig, db: str, include_auto_created: bool = False
    ) -> list[type[Model]]: ...
